from __future__ import annotations
import os
import sys
from pathlib import Path
from typing import BinaryIO
from sari.core.config import AppConfig
from sari.core.exceptions import ValidationError
from sari.core.models import ErrorResponseDTO
from sari.db.repositories.runtime_repository import RuntimeRepository
from sari.db.repositories.candidate_index_change_repository import CandidateIndexChangeRepository
from sari.db.repositories.symbol_cache_repository import SymbolCacheRepository
from sari.db.repositories.symbol_importance_repository import SymbolImportanceRepository
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.db.repositories.file_body_repository import FileBodyRepository
from sari.db.repositories.file_collection_repository import FileCollectionRepository
from sari.db.repositories.file_enrich_queue_repository import FileEnrichQueueRepository
from sari.db.repositories.lsp_tool_data_repository import LspToolDataRepository
from sari.db.repositories.knowledge_repository import KnowledgeRepository
from sari.db.repositories.vector_embedding_repository import VectorEmbeddingRepository
from sari.db.repositories.pipeline_benchmark_repository import PipelineBenchmarkRepository
from sari.db.repositories.pipeline_quality_repository import PipelineQualityRepository
from sari.db.repositories.language_probe_repository import LanguageProbeRepository
from sari.db.repositories.pipeline_lsp_matrix_repository import PipelineLspMatrixRepository
from sari.db.repositories.pipeline_control_state_repository import PipelineControlStateRepository
from sari.db.repositories.pipeline_job_event_repository import PipelineJobEventRepository
from sari.db.repositories.pipeline_error_event_repository import PipelineErrorEventRepository
from sari.db.repositories.pipeline_policy_repository import PipelinePolicyRepository
from sari.db.repositories.tool_readiness_repository import ToolReadinessRepository
from sari.db.migration import ensure_migrated
from sari.db.schema import init_schema
from sari.lsp.hub import LspHub
from sari.mcp.contracts import McpError, McpResponse
from sari.mcp.daemon_forward_policy import (
    StartDaemonFn,
    build_forward_error_message,
    default_start_daemon,
    extract_workspace_root,
    forward_with_retry,
    resolve_target,
    should_forward_to_daemon,
)
from sari.mcp.tools.admin_tools import DoctorTool, RepoCandidatesTool, RescanTool
from sari.mcp.tool_visibility import filter_tools_list_response_payload, is_hidden_tool_name
from sari.mcp.tools.pipeline_admin_tools import PipelineAutoSetTool, PipelineAutoStatusTool, PipelineAutoTickTool, PipelineAlertStatusTool, PipelineDeadListTool, PipelineDeadPurgeTool, PipelineDeadRequeueTool, PipelinePolicyGetTool, PipelinePolicySetTool
from sari.mcp.tools.pipeline_benchmark_tools import PipelineBenchmarkReportTool, PipelineBenchmarkRunTool
from sari.mcp.tools.pipeline_lsp_matrix_tools import PipelineLspMatrixReportTool, PipelineLspMatrixRunTool
from sari.mcp.tools.pipeline_quality_tools import PipelineQualityReportTool, PipelineQualityRunTool
from sari.mcp.tools.pack1 import pack1_error
from sari.mcp.tools.file_collection_tools import IndexFileTool, ListFilesTool, ReadFileTool, ScanOnceTool
from sari.mcp.tools.symbol_tools import GetCallersTool, SearchSymbolTool
from sari.mcp.transport import MCP_MODE_FRAMED, McpTransport, McpTransportParseError
from sari.mcp.server_daemon_forward import DaemonForwardError, forward_once
from sari.mcp.tools.search_tool import SearchTool
from sari.mcp.tools.legacy_tools import ArchiveContextTool, CallGraphHealthTool, CallGraphTool, DryRunDiffTool, GetContextTool, GetImplementationsTool, GetSnippetTool, KnowledgeTool, ListSymbolsTool, ReadSymbolTool, ReadTool, SariGuideTool, SaveSnippetTool, StatusTool
from sari.search.candidate_search import CandidateSearchService
from sari.search.hierarchy_scorer import HierarchyScorer
from sari.search.importance_scorer import ImportanceScorePolicyDTO, ImportanceScorer, ImportanceWeightsDTO
from sari.search.orchestrator import RankingBlendConfigDTO, SearchOrchestrator
from sari.search.symbol_resolve import SymbolResolveService
from sari.search.vector_reranker import VectorConfigDTO, VectorIndexSink, VectorReranker
from sari.services.admin_service import AdminService
from sari.services.file_collection_service import build_default_file_collection_service
from sari.services.pipeline_benchmark_service import BenchmarkLspExtractionBackend, PipelineBenchmarkService
from sari.services.pipeline_control_service import PipelineControlService
from sari.services.language_probe_service import LanguageProbeService
from sari.services.pipeline_lsp_matrix_service import PipelineLspMatrixService
from sari.services.pipeline_quality_service import PipelineQualityService, SerenaGoldenBackend

class McpServer:
    _DEFAULT_PROTOCOL_VERSION = '2024-11-05'
    _SUPPORTED_PROTOCOL_VERSIONS = (
        '2025-11-25',
        '2025-06-18',
        '2025-03-26',
        '2024-11-05',
    )

    def __init__(self, db_path: Path) -> None:
        init_schema(db_path)
        ensure_migrated(db_path)
        self._db_path = db_path
        self._proxy_to_daemon = os.getenv('SARI_MCP_FORWARD_TO_DAEMON', '').strip().lower() in {'1', 'true', 'yes', 'on'}
        self._daemon_autostart_on_failure = os.getenv('SARI_MCP_DAEMON_AUTOSTART', '1').strip().lower() in {'1', 'true', 'yes', 'on'}
        self._daemon_forward_timeout_sec = _parse_daemon_forward_timeout(os.getenv('SARI_MCP_DAEMON_TIMEOUT_SEC', '').strip())
        self._daemon_start_fn: StartDaemonFn = default_start_daemon
        runtime_config = AppConfig.default()
        workspace_repo = WorkspaceRepository(db_path)
        self._workspace_repo = workspace_repo
        runtime_repo = RuntimeRepository(db_path)
        self._runtime_repo = runtime_repo
        symbol_cache_repo = SymbolCacheRepository(db_path)
        symbol_importance_repo = SymbolImportanceRepository(db_path)
        file_repo = FileCollectionRepository(db_path)
        enrich_queue_repo = FileEnrichQueueRepository(db_path)
        body_repo = FileBodyRepository(db_path)
        lsp_repo = LspToolDataRepository(db_path)
        knowledge_repo = KnowledgeRepository(db_path)
        readiness_repo = ToolReadinessRepository(db_path)
        policy_repo = PipelinePolicyRepository(db_path)
        control_state_repo = PipelineControlStateRepository(db_path)
        event_repo = PipelineJobEventRepository(db_path)
        error_event_repo = PipelineErrorEventRepository(db_path)
        benchmark_repo = PipelineBenchmarkRepository(db_path)
        quality_repo = PipelineQualityRepository(db_path)
        language_probe_repo = LanguageProbeRepository(db_path)
        lsp_matrix_repo = PipelineLspMatrixRepository(db_path)
        vector_repo = VectorEmbeddingRepository(db_path)
        candidate_change_repo = CandidateIndexChangeRepository(db_path)
        importance_scorer = ImportanceScorer(file_repo=file_repo, lsp_repo=lsp_repo, cache_repo=symbol_importance_repo, weights=ImportanceWeightsDTO(kind_class=runtime_config.importance_kind_class, kind_function=runtime_config.importance_kind_function, kind_interface=runtime_config.importance_kind_interface, kind_method=runtime_config.importance_kind_method, fan_in_weight=runtime_config.importance_fan_in_weight, filename_exact_bonus=runtime_config.importance_filename_exact_bonus, core_path_bonus=runtime_config.importance_core_path_bonus, noisy_path_penalty=runtime_config.importance_noisy_path_penalty, code_ext_bonus=runtime_config.importance_code_ext_bonus, noisy_ext_penalty=runtime_config.importance_noisy_ext_penalty, recency_24h_multiplier=runtime_config.importance_recency_24h_multiplier, recency_7d_multiplier=runtime_config.importance_recency_7d_multiplier, recency_30d_multiplier=runtime_config.importance_recency_30d_multiplier), policy=ImportanceScorePolicyDTO(normalize_mode=runtime_config.importance_normalize_mode, max_importance_boost=runtime_config.importance_max_boost), core_path_tokens=runtime_config.importance_core_path_tokens, noisy_path_tokens=runtime_config.importance_noisy_path_tokens, code_extensions=runtime_config.importance_code_extensions, noisy_extensions=runtime_config.importance_noisy_extensions)
        vector_config = VectorConfigDTO(enabled=runtime_config.vector_enabled, model_id=runtime_config.vector_model_id, dim=runtime_config.vector_dim, candidate_k=runtime_config.vector_candidate_k, rerank_k=runtime_config.vector_rerank_k, blend_weight=runtime_config.vector_blend_weight, min_similarity_threshold=runtime_config.vector_min_similarity_threshold, max_vector_boost=runtime_config.vector_max_boost, min_token_count_for_rerank=runtime_config.vector_min_token_count_for_rerank, apply_to_item_types=runtime_config.vector_apply_to_item_types)
        vector_sink = VectorIndexSink(repository=vector_repo, config=vector_config)
        vector_reranker = VectorReranker(repository=vector_repo, config=vector_config)
        hierarchy_scorer = HierarchyScorer()
        candidate_service = CandidateSearchService.build_default(max_file_size_bytes=512 * 1024, index_root=db_path.parent / 'candidate_index', backend_mode='tantivy', enable_scan_fallback=True, change_repo=candidate_change_repo)
        file_collection_service = build_default_file_collection_service(workspace_repo=workspace_repo, file_repo=file_repo, enrich_queue_repo=enrich_queue_repo, body_repo=body_repo, lsp_repo=lsp_repo, readiness_repo=readiness_repo, policy_repo=policy_repo, event_repo=event_repo, error_event_repo=error_event_repo, candidate_index_sink=candidate_service, vector_index_sink=vector_sink, include_ext=runtime_config.collection_include_ext, exclude_globs=runtime_config.collection_exclude_globs, watcher_debounce_ms=runtime_config.watcher_debounce_ms, run_mode='prod')
        benchmark_collection_service = build_default_file_collection_service(workspace_repo=workspace_repo, file_repo=file_repo, enrich_queue_repo=enrich_queue_repo, body_repo=body_repo, lsp_repo=lsp_repo, readiness_repo=readiness_repo, policy_repo=policy_repo, event_repo=event_repo, error_event_repo=error_event_repo, run_mode='prod', lsp_backend=BenchmarkLspExtractionBackend(), persist_body_for_read=False)
        benchmark_service = PipelineBenchmarkService(file_collection_service=benchmark_collection_service, queue_repo=enrich_queue_repo, lsp_repo=lsp_repo, policy_repo=policy_repo, benchmark_repo=benchmark_repo, artifact_root=db_path.parent / 'artifacts')
        quality_service = PipelineQualityService(file_repo=file_repo, lsp_repo=lsp_repo, quality_repo=quality_repo, golden_backend=SerenaGoldenBackend(hub=LspHub(request_timeout_sec=runtime_config.lsp_request_timeout_sec)), artifact_root=db_path.parent / 'artifacts')
        language_probe_service = LanguageProbeService(workspace_repo=workspace_repo, lsp_hub=LspHub(request_timeout_sec=runtime_config.lsp_request_timeout_sec), probe_repo=language_probe_repo)
        pipeline_lsp_matrix_service = PipelineLspMatrixService(probe_service=language_probe_service, run_repo=lsp_matrix_repo)
        pipeline_control_service = PipelineControlService(policy_repo=policy_repo, event_repo=event_repo, queue_repo=enrich_queue_repo, control_state_repo=control_state_repo)
        admin_service = AdminService(config=AppConfig(db_path=db_path, host='127.0.0.1', preferred_port=47777, max_port_scan=50, stop_grace_sec=10), workspace_repo=workspace_repo, runtime_repo=runtime_repo, symbol_cache_repo=symbol_cache_repo)
        orchestrator = SearchOrchestrator(workspace_repo=workspace_repo, candidate_service=candidate_service, symbol_service=SymbolResolveService(hub=LspHub(request_timeout_sec=runtime_config.lsp_request_timeout_sec), cache_repo=symbol_cache_repo), importance_scorer=importance_scorer, hierarchy_scorer=hierarchy_scorer, vector_reranker=vector_reranker, blend_config=RankingBlendConfigDTO(w_rrf=runtime_config.ranking_w_rrf, w_importance=runtime_config.ranking_w_importance, w_vector=runtime_config.ranking_w_vector, w_hierarchy=runtime_config.ranking_w_hierarchy, version='v2-config'))
        self._search_tool = SearchTool(orchestrator=orchestrator, workspace_repo=workspace_repo, metrics_provider=file_collection_service.get_pipeline_metrics)
        self._doctor_tool = DoctorTool(admin_service=admin_service, workspace_repo=workspace_repo)
        self._rescan_tool = RescanTool(admin_service=admin_service, workspace_repo=workspace_repo)
        self._repo_candidates_tool = RepoCandidatesTool(admin_service=admin_service, workspace_repo=workspace_repo)
        self._scan_once_tool = ScanOnceTool(workspace_repo=workspace_repo, collection_service=file_collection_service)
        self._list_files_tool = ListFilesTool(workspace_repo=workspace_repo, collection_service=file_collection_service)
        self._read_file_tool = ReadFileTool(workspace_repo=workspace_repo, collection_service=file_collection_service)
        self._index_file_tool = IndexFileTool(workspace_repo=workspace_repo, collection_service=file_collection_service)
        self._search_symbol_tool = SearchSymbolTool(workspace_repo=workspace_repo, lsp_repo=lsp_repo)
        self._get_callers_tool = GetCallersTool(workspace_repo=workspace_repo, lsp_repo=lsp_repo)
        self._sari_guide_tool = SariGuideTool()
        self._status_tool = StatusTool(workspace_repo=workspace_repo, runtime_repo=runtime_repo, file_repo=file_repo, lsp_repo=lsp_repo, language_probe_repo=language_probe_repo)
        self._read_tool = ReadTool(workspace_repo=workspace_repo, file_collection_service=file_collection_service, lsp_repo=lsp_repo, knowledge_repo=knowledge_repo)
        self._dry_run_diff_tool = DryRunDiffTool(read_tool=self._read_tool)
        self._list_symbols_tool = ListSymbolsTool(workspace_repo=workspace_repo, lsp_repo=lsp_repo)
        self._read_symbol_tool = ReadSymbolTool(workspace_repo=workspace_repo, lsp_repo=lsp_repo)
        self._get_implementations_tool = GetImplementationsTool(workspace_repo=workspace_repo, lsp_repo=lsp_repo)
        self._call_graph_tool = CallGraphTool(workspace_repo=workspace_repo, lsp_repo=lsp_repo)
        self._call_graph_health_tool = CallGraphHealthTool(workspace_repo=workspace_repo, lsp_repo=lsp_repo)
        self._knowledge_tool = KnowledgeTool(workspace_repo=workspace_repo, knowledge_repo=knowledge_repo)
        self._save_snippet_tool = SaveSnippetTool(workspace_repo=workspace_repo, knowledge_repo=knowledge_repo)
        self._get_snippet_tool = GetSnippetTool(workspace_repo=workspace_repo, knowledge_repo=knowledge_repo)
        self._archive_context_tool = ArchiveContextTool(workspace_repo=workspace_repo, knowledge_repo=knowledge_repo)
        self._get_context_tool = GetContextTool(workspace_repo=workspace_repo, knowledge_repo=knowledge_repo)
        self._pipeline_policy_get_tool = PipelinePolicyGetTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_policy_set_tool = PipelinePolicySetTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_alert_status_tool = PipelineAlertStatusTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_dead_list_tool = PipelineDeadListTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_dead_requeue_tool = PipelineDeadRequeueTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_dead_purge_tool = PipelineDeadPurgeTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_auto_status_tool = PipelineAutoStatusTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_auto_set_tool = PipelineAutoSetTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_auto_tick_tool = PipelineAutoTickTool(workspace_repo=workspace_repo, service=pipeline_control_service)
        self._pipeline_benchmark_run_tool = PipelineBenchmarkRunTool(workspace_repo=workspace_repo, benchmark_service=benchmark_service)
        self._pipeline_benchmark_report_tool = PipelineBenchmarkReportTool(workspace_repo=workspace_repo, benchmark_service=benchmark_service)
        self._pipeline_quality_run_tool = PipelineQualityRunTool(workspace_repo=workspace_repo, quality_service=quality_service)
        self._pipeline_quality_report_tool = PipelineQualityReportTool(workspace_repo=workspace_repo, quality_service=quality_service)
        self._pipeline_lsp_matrix_run_tool = PipelineLspMatrixRunTool(workspace_repo=workspace_repo, matrix_service=pipeline_lsp_matrix_service)
        self._pipeline_lsp_matrix_report_tool = PipelineLspMatrixReportTool(workspace_repo=workspace_repo, matrix_service=pipeline_lsp_matrix_service)

    def handle_request(self, payload: dict[str, object]) -> McpResponse:
        request_id = payload.get('id')
        method = payload.get('method')
        if not isinstance(method, str):
            return McpResponse(request_id=request_id, result=None, error=McpError(code=-32600, message='invalid request'))
        if self._should_forward(payload, method):
            return self._forward_to_daemon(payload=payload, request_id=request_id)
        if method == 'initialize':
            params = payload.get('params')
            if not isinstance(params, dict):
                params = {}
            try:
                protocol_version = self._negotiate_protocol_version(params=params)
            except ValueError as exc:
                return McpResponse(request_id=request_id, result=None, error=McpError(code=-32602, message=str(exc)))
            return McpResponse(request_id=request_id, result={'protocolVersion': protocol_version, 'serverInfo': {'name': 'sari-v2', 'version': '0.1.0'}, 'capabilities': {'tools': {}}}, error=None)
        if method == 'sari/identify':
            return McpResponse(request_id=request_id, result={'name': 'sari-v2', 'version': '0.1.0', 'workspaceRoot': self._default_workspace_root(), 'pid': os.getpid()}, error=None)
        if method == 'prompts/list':
            return McpResponse(request_id=request_id, result={'prompts': []}, error=None)
        if method == 'resources/list':
            return McpResponse(request_id=request_id, result={'resources': []}, error=None)
        if method == 'resources/templates/list':
            return McpResponse(request_id=request_id, result={'resourceTemplates': []}, error=None)
        if method == 'roots/list':
            return McpResponse(request_id=request_id, result={'roots': self._roots_list()}, error=None)
        if method == 'initialized':
            return McpResponse(request_id=request_id, result={}, error=None)
        if method == 'notifications/initialized':
            return McpResponse(request_id=request_id, result={}, error=None)
        if method == 'ping':
            return McpResponse(request_id=request_id, result={}, error=None)
        if method == 'tools/list':
            return McpResponse(request_id=request_id, result={'tools': [{'name': 'sari_guide', 'description': 'Return quick usage guide in pack1 format', 'inputSchema': {'type': 'object', 'properties': {}}}, {'name': 'status', 'description': 'Return repository runtime/index status in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'search', 'description': 'Search symbols/files in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'query'], 'properties': {'repo': {'type': 'string'}, 'query': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'doctor', 'description': 'Return runtime health checks in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'rescan', 'description': 'Invalidate symbol cache in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'repo_candidates', 'description': 'List repository candidates in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'read', 'description': 'Unified read interface in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'mode': {'type': 'string'}, 'target': {'type': 'string'}, 'path': {'type': 'string'}, 'offset': {'type': 'integer', 'minimum': 0}, 'limit': {'type': 'integer', 'minimum': 1}, 'content': {'type': 'string'}, 'against': {'type': 'string'}, 'tag': {'type': 'string'}}}}, {'name': 'dry_run_diff', 'description': 'Legacy diff preview wrapper in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'path', 'content'], 'properties': {'repo': {'type': 'string'}, 'path': {'type': 'string'}, 'content': {'type': 'string'}, 'against': {'type': 'string'}}}}, {'name': 'scan_once', 'description': 'Scan repository files once and enqueue enrich jobs', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'list_files', 'description': 'List indexed files for repository in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}, 'prefix': {'type': 'string'}}}}, {'name': 'read_file', 'description': 'Read indexed file content in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'relative_path'], 'properties': {'repo': {'type': 'string'}, 'relative_path': {'type': 'string'}, 'offset': {'type': 'integer', 'minimum': 0}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'index_file', 'description': 'Incrementally index single file in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'relative_path'], 'properties': {'repo': {'type': 'string'}, 'relative_path': {'type': 'string'}}}}, {'name': 'list_symbols', 'description': 'List indexed symbols in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'query': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'read_symbol', 'description': 'Read symbol detail in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'name': {'type': 'string'}, 'symbol_id': {'type': 'string'}, 'sid': {'type': 'string'}, 'path': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'search_symbol', 'description': 'Search indexed symbols in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'query'], 'properties': {'repo': {'type': 'string'}, 'query': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}, 'path_prefix': {'type': 'string'}}}}, {'name': 'get_callers', 'description': 'Get caller edges for symbol in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'symbol': {'type': 'string'}, 'symbol_id': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'get_implementations', 'description': 'Get implementation candidates for symbol in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'symbol': {'type': 'string'}, 'symbol_id': {'type': 'string'}, 'sid': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'call_graph', 'description': 'Get call graph for symbol in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'symbol': {'type': 'string'}, 'symbol_id': {'type': 'string'}, 'sid': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'call_graph_health', 'description': 'Get call graph health summary in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'knowledge', 'description': 'Query archived knowledge entries in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'query': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'save_snippet', 'description': 'Save code snippet to local store in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'path', 'start_line', 'end_line', 'tag'], 'properties': {'repo': {'type': 'string'}, 'path': {'type': 'string'}, 'start_line': {'type': 'integer', 'minimum': 1}, 'end_line': {'type': 'integer', 'minimum': 1}, 'tag': {'type': 'string'}, 'note': {'type': 'string'}, 'commit': {'type': 'string'}}}}, {'name': 'get_snippet', 'description': 'Query saved snippets in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'tag': {'type': 'string'}, 'query': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'archive_context', 'description': 'Archive context notes in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'topic', 'content'], 'properties': {'repo': {'type': 'string'}, 'topic': {'type': 'string'}, 'content': {'type': 'string'}, 'tags': {'type': 'array', 'items': {'type': 'string'}}, 'related_files': {'type': 'array', 'items': {'type': 'string'}}}}}, {'name': 'get_context', 'description': 'Get archived context notes in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'query': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'pipeline_policy_get', 'description': 'Get pipeline policy in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'pipeline_policy_set', 'description': 'Set pipeline policy in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'deletion_hold': {'type': ['string', 'boolean']}, 'l3_p95_threshold_ms': {'type': 'integer', 'minimum': 1}, 'dead_ratio_threshold_bps': {'type': 'integer', 'minimum': 1}, 'workers': {'type': 'integer', 'minimum': 1}, 'alert_window_sec': {'type': 'integer', 'minimum': 60}}}}, {'name': 'pipeline_alert_status', 'description': 'Get pipeline alert snapshot in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'pipeline_dead_list', 'description': 'List dead enrich jobs in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}}}}, {'name': 'pipeline_dead_requeue', 'description': 'Requeue dead enrich jobs in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}, 'all': {'type': ['boolean', 'string']}}}}, {'name': 'pipeline_dead_purge', 'description': 'Purge dead enrich jobs in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'limit': {'type': 'integer', 'minimum': 1}, 'all': {'type': ['boolean', 'string']}}}}, {'name': 'pipeline_auto_status', 'description': 'Get pipeline auto-control state in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'pipeline_auto_set', 'description': 'Set pipeline auto-control enabled flag in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo', 'enabled'], 'properties': {'repo': {'type': 'string'}, 'enabled': {'type': ['string', 'boolean']}}}}, {'name': 'pipeline_auto_tick', 'description': 'Evaluate pipeline auto-control once in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'pipeline_benchmark_run', 'description': 'Run pipeline benchmark and return summary in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'target_files': {'type': 'integer', 'minimum': 1}, 'profile': {'type': 'string'}, 'language_filter': {'type': ['string', 'array'], 'items': {'type': 'string'}}, 'per_language_report': {'type': ['boolean', 'string']}}}}, {'name': 'pipeline_benchmark_report', 'description': 'Return latest pipeline benchmark summary in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'pipeline_quality_run', 'description': 'Run L3 quality evaluation in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'limit_files': {'type': 'integer', 'minimum': 1}, 'profile': {'type': 'string'}, 'language_filter': {'type': ['string', 'array'], 'items': {'type': 'string'}}}}}, {'name': 'pipeline_quality_report', 'description': 'Return latest L3 quality summary in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}, {'name': 'pipeline_lsp_matrix_run', 'description': 'Run LSP readiness matrix and hard gate in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}, 'required_languages': {'type': ['string', 'array'], 'items': {'type': 'string'}}, 'fail_on_unavailable': {'type': ['boolean', 'string']}, 'strict_all_languages': {'type': ['boolean', 'string']}, 'strict_symbol_gate': {'type': ['boolean', 'string']}}}}, {'name': 'pipeline_lsp_matrix_report', 'description': 'Return latest LSP readiness matrix report in pack1 format', 'inputSchema': {'type': 'object', 'required': ['repo'], 'properties': {'repo': {'type': 'string'}}}}]}, error=None)
        if method == 'tools/call':
            params = payload.get('params')
            if not isinstance(params, dict):
                return McpResponse(request_id=request_id, result=None, error=McpError(code=-32602, message='invalid params'))
            tool_name = params.get('name')
            if is_hidden_tool_name(tool_name):
                return McpResponse(request_id=request_id, result=None, error=McpError(code=-32601, message='tool not found'))
            arguments = params.get('arguments', {})
            if not isinstance(arguments, dict):
                return McpResponse(request_id=request_id, result=None, error=McpError(code=-32602, message='invalid arguments'))
            try:
                if tool_name == 'search':
                    return McpResponse(request_id=request_id, result=self._search_tool.call(arguments), error=None)
                if tool_name == 'sari_guide':
                    return McpResponse(request_id=request_id, result=self._sari_guide_tool.call(arguments), error=None)
                if tool_name == 'status':
                    return McpResponse(request_id=request_id, result=self._status_tool.call(arguments), error=None)
                if tool_name == 'doctor':
                    return McpResponse(request_id=request_id, result=self._doctor_tool.call(arguments), error=None)
                if tool_name == 'rescan':
                    return McpResponse(request_id=request_id, result=self._rescan_tool.call(arguments), error=None)
                if tool_name == 'repo_candidates':
                    return McpResponse(request_id=request_id, result=self._repo_candidates_tool.call(arguments), error=None)
                if tool_name == 'read':
                    return McpResponse(request_id=request_id, result=self._read_tool.call(arguments), error=None)
                if tool_name == 'dry_run_diff':
                    return McpResponse(request_id=request_id, result=self._dry_run_diff_tool.call(arguments), error=None)
                if tool_name == 'scan_once':
                    return McpResponse(request_id=request_id, result=self._scan_once_tool.call(arguments), error=None)
                if tool_name == 'list_files':
                    return McpResponse(request_id=request_id, result=self._list_files_tool.call(arguments), error=None)
                if tool_name == 'read_file':
                    return McpResponse(request_id=request_id, result=self._read_file_tool.call(arguments), error=None)
                if tool_name == 'index_file':
                    return McpResponse(request_id=request_id, result=self._index_file_tool.call(arguments), error=None)
                if tool_name == 'list_symbols':
                    return McpResponse(request_id=request_id, result=self._list_symbols_tool.call(arguments), error=None)
                if tool_name == 'read_symbol':
                    return McpResponse(request_id=request_id, result=self._read_symbol_tool.call(arguments), error=None)
                if tool_name == 'search_symbol':
                    return McpResponse(request_id=request_id, result=self._search_symbol_tool.call(arguments), error=None)
                if tool_name == 'get_callers':
                    return McpResponse(request_id=request_id, result=self._get_callers_tool.call(arguments), error=None)
                if tool_name == 'get_implementations':
                    return McpResponse(request_id=request_id, result=self._get_implementations_tool.call(arguments), error=None)
                if tool_name == 'call_graph':
                    return McpResponse(request_id=request_id, result=self._call_graph_tool.call(arguments), error=None)
                if tool_name == 'call_graph_health':
                    return McpResponse(request_id=request_id, result=self._call_graph_health_tool.call(arguments), error=None)
                if tool_name == 'knowledge':
                    return McpResponse(request_id=request_id, result=self._knowledge_tool.call(arguments), error=None)
                if tool_name == 'save_snippet':
                    return McpResponse(request_id=request_id, result=self._save_snippet_tool.call(arguments), error=None)
                if tool_name == 'get_snippet':
                    return McpResponse(request_id=request_id, result=self._get_snippet_tool.call(arguments), error=None)
                if tool_name == 'archive_context':
                    return McpResponse(request_id=request_id, result=self._archive_context_tool.call(arguments), error=None)
                if tool_name == 'get_context':
                    return McpResponse(request_id=request_id, result=self._get_context_tool.call(arguments), error=None)
                if tool_name == 'pipeline_policy_get':
                    return McpResponse(request_id=request_id, result=self._pipeline_policy_get_tool.call(arguments), error=None)
                if tool_name == 'pipeline_policy_set':
                    return McpResponse(request_id=request_id, result=self._pipeline_policy_set_tool.call(arguments), error=None)
                if tool_name == 'pipeline_alert_status':
                    return McpResponse(request_id=request_id, result=self._pipeline_alert_status_tool.call(arguments), error=None)
                if tool_name == 'pipeline_dead_list':
                    return McpResponse(request_id=request_id, result=self._pipeline_dead_list_tool.call(arguments), error=None)
                if tool_name == 'pipeline_dead_requeue':
                    return McpResponse(request_id=request_id, result=self._pipeline_dead_requeue_tool.call(arguments), error=None)
                if tool_name == 'pipeline_dead_purge':
                    return McpResponse(request_id=request_id, result=self._pipeline_dead_purge_tool.call(arguments), error=None)
                if tool_name == 'pipeline_auto_status':
                    return McpResponse(request_id=request_id, result=self._pipeline_auto_status_tool.call(arguments), error=None)
                if tool_name == 'pipeline_auto_set':
                    return McpResponse(request_id=request_id, result=self._pipeline_auto_set_tool.call(arguments), error=None)
                if tool_name == 'pipeline_auto_tick':
                    return McpResponse(request_id=request_id, result=self._pipeline_auto_tick_tool.call(arguments), error=None)
                if tool_name == 'pipeline_benchmark_run':
                    return McpResponse(request_id=request_id, result=self._pipeline_benchmark_run_tool.call(arguments), error=None)
                if tool_name == 'pipeline_benchmark_report':
                    return McpResponse(request_id=request_id, result=self._pipeline_benchmark_report_tool.call(arguments), error=None)
                if tool_name == 'pipeline_quality_run':
                    return McpResponse(request_id=request_id, result=self._pipeline_quality_run_tool.call(arguments), error=None)
                if tool_name == 'pipeline_quality_report':
                    return McpResponse(request_id=request_id, result=self._pipeline_quality_report_tool.call(arguments), error=None)
                if tool_name == 'pipeline_lsp_matrix_run':
                    return McpResponse(request_id=request_id, result=self._pipeline_lsp_matrix_run_tool.call(arguments), error=None)
                if tool_name == 'pipeline_lsp_matrix_report':
                    return McpResponse(request_id=request_id, result=self._pipeline_lsp_matrix_report_tool.call(arguments), error=None)
            except ValidationError as exc:
                payload = pack1_error(ErrorResponseDTO(code=exc.context.code, message=exc.context.message))
                return McpResponse(request_id=request_id, result=payload, error=None)
            return McpResponse(request_id=request_id, result=None, error=McpError(code=-32601, message='tool not found'))
        return McpResponse(request_id=request_id, result=None, error=McpError(code=-32601, message='method not found'))

    def _should_forward(self, payload: dict[str, object], method: str) -> bool:
        _ = payload
        return should_forward_to_daemon(proxy_enabled=self._proxy_to_daemon, method=method)

    def _forward_to_daemon(self, payload: dict[str, object], request_id: object) -> McpResponse:
        workspace_root = extract_workspace_root(payload)
        try:
            forwarded = forward_with_retry(
                request=payload,
                db_path=self._db_path,
                workspace_root=workspace_root,
                host_override=None,
                port_override=None,
                timeout_sec=self._daemon_forward_timeout_sec,
                auto_start_on_failure=self._daemon_autostart_on_failure,
                start_daemon_fn=self._daemon_start_fn,
                resolve_target_fn=resolve_target,
                forward_once_fn=forward_once,
            )
        except (OSError, TimeoutError, DaemonForwardError, ValueError) as exc:
            return McpResponse(request_id=request_id, result=None, error=McpError(code=-32002, message=build_forward_error_message(exc)))
        response_id = forwarded.get('id', request_id)
        error_payload = forwarded.get('error')
        if isinstance(error_payload, dict):
            code = error_payload.get('code')
            message = error_payload.get('message')
            if isinstance(code, int) and isinstance(message, str):
                return McpResponse(request_id=response_id, result=None, error=McpError(code=code, message=message))
            return McpResponse(request_id=response_id, result=None, error=McpError(code=-32003, message='invalid daemon error response'))
        return McpResponse(request_id=response_id, result=forwarded.get('result'), error=None)

    def _roots_list(self) -> list[dict[str, str]]:
        roots: list[dict[str, str]] = []
        for workspace in self._workspace_repo.list_all():
            root_path = workspace.path
            name = Path(root_path).name if Path(root_path).name != '' else root_path
            roots.append({'uri': f'file://{root_path}', 'name': name})
        return roots

    def _default_workspace_root(self) -> str:
        workspaces = self._workspace_repo.list_all()
        if len(workspaces) == 0:
            return ''
        return workspaces[0].path

    def _negotiate_protocol_version(self, params: dict[str, object]) -> str:
        versions = self._iter_client_protocol_versions(params=params)
        for candidate in versions:
            if candidate in self._SUPPORTED_PROTOCOL_VERSIONS:
                return candidate
        strict = os.getenv('SARI_STRICT_PROTOCOL', '').strip().lower() in {'1', 'true', 'yes', 'on'}
        if strict and len(versions) > 0:
            raise ValueError('Unsupported protocol version')
        return self._DEFAULT_PROTOCOL_VERSION

    def _iter_client_protocol_versions(self, params: dict[str, object]) -> list[str]:
        versions: list[str] = []
        seen: set[str] = set()

        def _append(raw_value: object) -> None:
            if not isinstance(raw_value, str):
                return
            normalized = raw_value.strip()
            if normalized == '' or normalized in seen:
                return
            seen.add(normalized)
            versions.append(normalized)

        _append(params.get('protocolVersion'))
        supported = params.get('supportedProtocolVersions')
        if isinstance(supported, list):
            for item in supported:
                _append(item)
        capabilities = params.get('capabilities')
        if isinstance(capabilities, dict):
            cap_versions = capabilities.get('protocolVersions')
            if isinstance(cap_versions, list):
                for item in cap_versions:
                    _append(item)
        return versions

def run_stdio_streams(db_path: Path, input_stream: BinaryIO, output_stream: BinaryIO) -> int:
    server = McpServer(db_path=db_path)
    runtime = server._runtime_repo.get_runtime()
    if runtime is not None:
        server._runtime_repo.increment_session()
    transport = McpTransport(input_stream=input_stream, output_stream=output_stream, allow_jsonl=True)
    transport.default_mode = MCP_MODE_FRAMED
    try:
        while True:
            try:
                read_result = transport.read_message()
            except McpTransportParseError as exc:
                parse_response = McpResponse(request_id=None, result=None, error=McpError(code=-32700, message=str(exc)))
                transport.write_message(parse_response.to_dict(), mode=exc.mode)
                continue
            if read_result is None:
                return 0
            payload, mode = read_result
            response = server.handle_request(payload)
            response_payload = response.to_dict()
            if str(payload.get("method", "")).strip() == "tools/list":
                response_payload = filter_tools_list_response_payload(response_payload)
            transport.write_message(response_payload, mode=mode)
    finally:
        runtime = server._runtime_repo.get_runtime()
        if runtime is not None:
            server._runtime_repo.decrement_session()
    return 0

def run_stdio(db_path: Path) -> int:
    input_stream = getattr(sys.stdin, 'buffer', sys.stdin)
    output_stream = getattr(sys.stdout, 'buffer', sys.stdout)
    return run_stdio_streams(db_path=db_path, input_stream=input_stream, output_stream=output_stream)

def _parse_daemon_forward_timeout(raw_value: str) -> float:
    if raw_value == '':
        return 2.0
    try:
        parsed = float(raw_value)
    except ValueError:
        return 2.0
    if parsed <= 0:
        return 2.0
    return parsed



"""검색 점수 결합 책임을 분리한 모듈."""

from __future__ import annotations

from sari.core.models import RankingComponentsDTO, SearchItemDTO


class ScoreBlender:
    """RRF 융합과 최종 점수 결합을 담당한다."""

    def __init__(self, rrf_k: int, w_rrf: float, w_importance: float, w_vector: float, w_hierarchy: float) -> None:
        """점수 결합 정책 파라미터를 초기화한다."""
        self._rrf_k = rrf_k
        self._w_rrf = w_rrf
        self._w_importance = w_importance
        self._w_vector = w_vector
        self._w_hierarchy = w_hierarchy

    def fuse_rrf(
        self,
        candidate_items: list[SearchItemDTO],
        resolved_items: list[SearchItemDTO],
        limit: int,
    ) -> list[SearchItemDTO]:
        """후보와 심볼 결과를 RRF로 융합한다."""
        score_map: dict[str, float] = {}
        item_map: dict[str, SearchItemDTO] = {}

        for rank, item in enumerate(candidate_items, start=1):
            key = self._item_key(item)
            item_map[key] = item
            score_map[key] = score_map.get(key, 0.0) + (1.0 / float(self._rrf_k + rank))

        for rank, item in enumerate(resolved_items, start=1):
            key = self._item_key(item)
            item_map[key] = item
            score_map[key] = score_map.get(key, 0.0) + (1.0 / float(self._rrf_k + rank))

        sorted_keys = sorted(score_map.keys(), key=lambda key: score_map[key], reverse=True)
        output: list[SearchItemDTO] = []
        for key in sorted_keys[:limit]:
            raw_item = item_map[key]
            output.append(
                SearchItemDTO(
                    item_type=raw_item.item_type,
                    repo=raw_item.repo,
                    relative_path=raw_item.relative_path,
                    score=score_map[key],
                    source=raw_item.source,
                    name=raw_item.name,
                    kind=raw_item.kind,
                    content_hash=raw_item.content_hash,
                    rrf_score=score_map[key],
                    importance_score=raw_item.importance_score,
                    base_rrf_score=score_map[key],
                    importance_norm_score=0.0,
                    vector_norm_score=0.0,
                    hierarchy_score=raw_item.hierarchy_score,
                    hierarchy_norm_score=0.0,
                    symbol_key=raw_item.symbol_key,
                    parent_symbol_key=raw_item.parent_symbol_key,
                    depth=raw_item.depth,
                    container_name=raw_item.container_name,
                    ranking_components=None,
                    vector_score=raw_item.vector_score,
                    blended_score=score_map[key],
                    final_score=score_map[key],
                )
            )
        return output

    def blend(self, items: list[SearchItemDTO], limit: int) -> list[SearchItemDTO]:
        """RRF/importance/vector를 정규화해 단일 점수로 결합한다."""
        if len(items) == 0:
            return []
        rrf_values = [item.rrf_score if item.rrf_score > 0.0 else item.score for item in items]
        importance_values = [item.importance_score for item in items]
        vector_values = [item.vector_score if item.vector_score is not None else 0.0 for item in items]
        hierarchy_values = [item.hierarchy_score for item in items]
        rrf_norm = self._normalize_minmax(rrf_values)
        importance_norm = self._normalize_minmax(importance_values)
        vector_norm = self._normalize_minmax(vector_values)
        hierarchy_norm = self._normalize_minmax(hierarchy_values)

        blended: list[SearchItemDTO] = []
        for index, item in enumerate(items):
            score = (
                (self._w_rrf * rrf_norm[index])
                + (self._w_importance * importance_norm[index])
                + (self._w_vector * vector_norm[index])
                + (self._w_hierarchy * hierarchy_norm[index])
            )
            blended.append(
                SearchItemDTO(
                    item_type=item.item_type,
                    repo=item.repo,
                    relative_path=item.relative_path,
                    score=score,
                    source=item.source,
                    name=item.name,
                    kind=item.kind,
                    content_hash=item.content_hash,
                    rrf_score=rrf_values[index],
                    importance_score=item.importance_score,
                    base_rrf_score=rrf_values[index],
                    importance_norm_score=importance_norm[index],
                    vector_norm_score=vector_norm[index],
                    hierarchy_score=item.hierarchy_score,
                    hierarchy_norm_score=hierarchy_norm[index],
                    symbol_key=item.symbol_key,
                    parent_symbol_key=item.parent_symbol_key,
                    depth=item.depth,
                    container_name=item.container_name,
                    ranking_components=RankingComponentsDTO(
                        rrf=rrf_norm[index],
                        importance=importance_norm[index],
                        vector=vector_norm[index],
                        hierarchy=hierarchy_norm[index],
                        final=score,
                    ),
                    vector_score=item.vector_score,
                    blended_score=score,
                    final_score=score,
                )
            )
        blended.sort(key=lambda item: item.score, reverse=True)
        return blended[:limit]

    def _normalize_minmax(self, values: list[float]) -> list[float]:
        """0..1 범위 min-max 정규화를 수행한다."""
        if len(values) == 0:
            return []
        minimum = min(values)
        maximum = max(values)
        if minimum == maximum:
            return [0.0 for _ in values]
        scale = maximum - minimum
        return [float((value - minimum) / scale) for value in values]

    def _item_key(self, item: SearchItemDTO) -> str:
        """RRF 융합용 아이템 고유 키를 계산한다."""
        if item.item_type == "file":
            return f"file:{item.repo}:{item.relative_path}"
        if item.symbol_key is not None and item.symbol_key.strip() != "":
            return f"symbol-key:{item.repo}:{item.symbol_key}"
        return f"symbol:{item.repo}:{item.relative_path}:{item.name}:{item.kind}"



"""HTTP/MCP 공용 read 실행 계층을 제공한다."""

from __future__ import annotations

from sari.db.repositories.knowledge_repository import KnowledgeRepository
from sari.db.repositories.lsp_tool_data_repository import LspToolDataRepository
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.mcp.tools.legacy_tools import ReadTool
from sari.services.collection.ports import CollectionScanPort


class ReadFacadeService:
    """MCP read 로직을 HTTP에서도 재사용하기 위한 파사드다."""

    def __init__(
        self,
        workspace_repo: WorkspaceRepository,
        file_collection_service: CollectionScanPort,
        lsp_repo: LspToolDataRepository,
        knowledge_repo: KnowledgeRepository,
    ) -> None:
        """ReadTool 의존성을 구성한다."""
        self._read_tool = ReadTool(
            workspace_repo=workspace_repo,
            file_collection_service=file_collection_service,
            lsp_repo=lsp_repo,
            knowledge_repo=knowledge_repo,
        )

    def read(self, arguments: dict[str, object]) -> dict[str, object]:
        """입력 인자로 read 결과(pack1)를 반환한다."""
        return self._read_tool.call(arguments=arguments)



"""L2 본문 보강 큐 저장소를 구현한다."""

from __future__ import annotations

import uuid
from datetime import datetime, timedelta
from pathlib import Path

from sari.core.models import DeadJobItemDTO, EnqueueRequestDTO, FileEnrichFailureUpdateDTO, FileEnrichJobDTO
from sari.db.row_mapper import row_int, row_optional_str, row_str
from sari.db.schema import connect


class FileEnrichQueueRepository:
    """L2 본문 보강 큐 영속화를 담당한다."""

    def __init__(self, db_path: Path) -> None:
        """저장소에 사용할 DB 경로를 저장한다."""
        self._db_path = db_path

    def enqueue(
        self,
        repo_root: str,
        relative_path: str,
        content_hash: str,
        priority: int,
        enqueue_source: str,
        now_iso: str,
    ) -> str:
        """L2 본문 보강 큐 작업을 적재한다."""
        job_id = str(uuid.uuid4())
        job = FileEnrichJobDTO(
            job_id=job_id,
            repo_root=repo_root,
            relative_path=relative_path,
            content_hash=content_hash,
            priority=priority,
            enqueue_source=enqueue_source,
            status="PENDING",
            attempt_count=0,
            last_error=None,
            next_retry_at=now_iso,
            created_at=now_iso,
            updated_at=now_iso,
        )
        with connect(self._db_path) as conn:
            existing = conn.execute(
                """
                SELECT job_id, priority
                FROM file_enrich_queue
                WHERE repo_root = :repo_root
                  AND relative_path = :relative_path
                  AND status IN ('PENDING', 'FAILED')
                ORDER BY updated_at DESC
                LIMIT 1
                """,
                {"repo_root": repo_root, "relative_path": relative_path},
            ).fetchone()
            if existing is not None:
                existing_job_id = row_str(existing, "job_id")
                merged_priority = max(row_int(existing, "priority"), priority)
                conn.execute(
                    """
                    UPDATE file_enrich_queue
                    SET content_hash = :content_hash,
                        priority = :priority,
                        enqueue_source = :enqueue_source,
                        status = 'PENDING',
                        next_retry_at = :next_retry_at,
                        updated_at = :updated_at,
                        last_error = NULL
                    WHERE job_id = :job_id
                    """,
                    {
                        "job_id": existing_job_id,
                        "content_hash": content_hash,
                        "priority": merged_priority,
                        "enqueue_source": enqueue_source,
                        "next_retry_at": now_iso,
                        "updated_at": now_iso,
                    },
                )
                conn.commit()
                return existing_job_id
            conn.execute(
                """
                INSERT INTO file_enrich_queue(
                    job_id, repo_root, relative_path, content_hash, content_raw, content_encoding,
                    priority, enqueue_source, status, attempt_count, last_error, next_retry_at, created_at, updated_at
                )
                VALUES(
                    :job_id, :repo_root, :relative_path, :content_hash, '', 'utf-8',
                    :priority, :enqueue_source, :status, :attempt_count, :last_error, :next_retry_at, :created_at, :updated_at
                )
                """,
                job.to_sql_params(),
            )
            conn.commit()
        return job_id

    def enqueue_many(self, requests: list[EnqueueRequestDTO]) -> list[str]:
        """L2 본문 보강 큐 작업을 배치 적재한다."""
        if len(requests) == 0:
            return []
        enqueued_ids: list[str] = []
        with connect(self._db_path) as conn:
            for request in requests:
                job_id = str(uuid.uuid4())
                job = FileEnrichJobDTO(
                    job_id=job_id,
                    repo_root=request.repo_root,
                    relative_path=request.relative_path,
                    content_hash=request.content_hash,
                    priority=request.priority,
                    enqueue_source=request.enqueue_source,
                    status="PENDING",
                    attempt_count=0,
                    last_error=None,
                    next_retry_at=request.now_iso,
                    created_at=request.now_iso,
                    updated_at=request.now_iso,
                )
                existing = conn.execute(
                    """
                    SELECT job_id, priority
                    FROM file_enrich_queue
                    WHERE repo_root = :repo_root
                      AND relative_path = :relative_path
                      AND status IN ('PENDING', 'FAILED')
                    ORDER BY updated_at DESC
                    LIMIT 1
                    """,
                    {"repo_root": request.repo_root, "relative_path": request.relative_path},
                ).fetchone()
                if existing is not None:
                    existing_job_id = row_str(existing, "job_id")
                    merged_priority = max(row_int(existing, "priority"), request.priority)
                    conn.execute(
                        """
                        UPDATE file_enrich_queue
                        SET content_hash = :content_hash,
                            priority = :priority,
                            enqueue_source = :enqueue_source,
                            status = 'PENDING',
                            next_retry_at = :next_retry_at,
                            updated_at = :updated_at,
                            last_error = NULL
                        WHERE job_id = :job_id
                        """,
                        {
                            "job_id": existing_job_id,
                            "content_hash": request.content_hash,
                            "priority": merged_priority,
                            "enqueue_source": request.enqueue_source,
                            "next_retry_at": request.now_iso,
                            "updated_at": request.now_iso,
                        },
                    )
                    enqueued_ids.append(existing_job_id)
                    continue
                conn.execute(
                    """
                    INSERT INTO file_enrich_queue(
                        job_id, repo_root, relative_path, content_hash, content_raw, content_encoding,
                        priority, enqueue_source, status, attempt_count, last_error, next_retry_at, created_at, updated_at
                    )
                    VALUES(
                        :job_id, :repo_root, :relative_path, :content_hash, '', 'utf-8',
                        :priority, :enqueue_source, :status, :attempt_count, :last_error, :next_retry_at, :created_at, :updated_at
                    )
                    """,
                    job.to_sql_params(),
                )
                enqueued_ids.append(job_id)
            conn.commit()
        return enqueued_ids

    def acquire_pending(self, limit: int, now_iso: str) -> list[FileEnrichJobDTO]:
        """처리 가능한 보강 작업을 RUNNING으로 전환하고 반환한다."""
        return self._acquire_pending_internal(limit=limit, now_iso=now_iso, source_mode="all")

    def acquire_pending_for_l2(self, limit: int, now_iso: str) -> list[FileEnrichJobDTO]:
        """L2 단계에서 처리할 작업을 RUNNING으로 전환하고 반환한다."""
        return self._acquire_pending_internal(limit=limit, now_iso=now_iso, source_mode="l2")

    def acquire_pending_for_l3(self, limit: int, now_iso: str) -> list[FileEnrichJobDTO]:
        """L3 단계에서 처리할 작업을 RUNNING으로 전환하고 반환한다."""
        return self._acquire_pending_internal(limit=limit, now_iso=now_iso, source_mode="l3")

    def _acquire_pending_internal(self, limit: int, now_iso: str, source_mode: str) -> list[FileEnrichJobDTO]:
        """단계 소스 조건에 맞는 보강 작업을 RUNNING으로 전환하고 반환한다."""
        source_clause = ""
        if source_mode == "l2":
            source_clause = "AND enqueue_source <> 'l3'"
        elif source_mode == "l3":
            source_clause = "AND enqueue_source = 'l3'"
        with connect(self._db_path) as conn:
            rows = conn.execute(
                f"""
                WITH picked AS (
                    SELECT job_id
                    FROM file_enrich_queue
                    WHERE status IN ('PENDING', 'FAILED')
                      AND next_retry_at <= :now_iso
                      {source_clause}
                    ORDER BY priority DESC, next_retry_at ASC, created_at ASC
                    LIMIT :limit
                )
                UPDATE file_enrich_queue
                SET status = 'RUNNING',
                    updated_at = :now_iso
                WHERE job_id IN (SELECT job_id FROM picked)
                  AND status IN ('PENDING', 'FAILED')
                  AND next_retry_at <= :now_iso
                RETURNING job_id, repo_root, relative_path, content_hash, priority, enqueue_source,
                          status, attempt_count, last_error, next_retry_at, created_at, updated_at
                """,
                {"limit": limit, "now_iso": now_iso},
            ).fetchall()

            jobs: list[FileEnrichJobDTO] = []
            for row in rows:
                jobs.append(
                    FileEnrichJobDTO(
                        job_id=row_str(row, "job_id"),
                        repo_root=row_str(row, "repo_root"),
                        relative_path=row_str(row, "relative_path"),
                        content_hash=row_str(row, "content_hash"),
                        priority=row_int(row, "priority"),
                        enqueue_source=row_str(row, "enqueue_source"),
                        status=row_str(row, "status"),
                        attempt_count=row_int(row, "attempt_count"),
                        last_error=row_optional_str(row, "last_error"),
                        next_retry_at=row_str(row, "next_retry_at"),
                        created_at=row_str(row, "created_at"),
                        updated_at=row_str(row, "updated_at"),
                    )
                )
            jobs.sort(key=lambda item: (-item.priority, item.next_retry_at, item.created_at))
            conn.commit()
            return jobs

    def promote_to_l3(self, job_id: str, now_iso: str) -> None:
        """L2 완료 작업을 L3 대기 상태로 승격한다."""
        self.promote_to_l3_many(job_ids=[job_id], now_iso=now_iso)

    def promote_to_l3_many(self, job_ids: list[str], now_iso: str) -> None:
        """L2 완료 작업들을 L3 대기 상태로 배치 승격한다."""
        if len(job_ids) == 0:
            return
        with connect(self._db_path) as conn:
            for job_id in job_ids:
                conn.execute(
                    """
                    UPDATE file_enrich_queue
                    SET status = 'PENDING',
                        enqueue_source = 'l3',
                        attempt_count = 0,
                        last_error = NULL,
                        next_retry_at = :now_iso,
                        updated_at = :now_iso
                    WHERE job_id = :job_id
                    """,
                    {"job_id": job_id, "now_iso": now_iso},
                )
            conn.commit()

    def mark_done(self, job_id: str) -> None:
        """보강 작업을 완료 상태로 갱신한다."""
        with connect(self._db_path) as conn:
            conn.execute("UPDATE file_enrich_queue SET status = 'DONE' WHERE job_id = :job_id", {"job_id": job_id})
            conn.commit()

    def mark_done_many(self, job_ids: list[str]) -> None:
        """보강 작업 완료 상태를 배치로 갱신한다."""
        if len(job_ids) == 0:
            return
        with connect(self._db_path) as conn:
            for job_id in job_ids:
                conn.execute("UPDATE file_enrich_queue SET status = 'DONE' WHERE job_id = :job_id", {"job_id": job_id})
            conn.commit()

    def mark_failed(self, job_id: str, error_message: str, next_retry_at: str, dead_threshold: int) -> None:
        """보강 작업 실패 상태를 갱신한다."""
        self.mark_failed_with_backoff(
            job_id=job_id,
            error_message=error_message,
            now_iso=next_retry_at,
            dead_threshold=dead_threshold,
            backoff_base_sec=1,
        )

    def mark_failed_with_backoff(
        self,
        job_id: str,
        error_message: str,
        now_iso: str,
        dead_threshold: int,
        backoff_base_sec: int,
    ) -> None:
        """보강 작업 실패 상태를 백오프 정책으로 갱신한다."""
        with connect(self._db_path) as conn:
            row = conn.execute(
                "SELECT attempt_count FROM file_enrich_queue WHERE job_id = :job_id",
                {"job_id": job_id},
            ).fetchone()
            if row is None:
                return
            attempt_count = row_int(row, "attempt_count") + 1
            status = "DEAD" if attempt_count >= dead_threshold else "FAILED"
            next_retry_at = self._compute_next_retry_at(now_iso=now_iso, attempt_count=attempt_count, backoff_base_sec=backoff_base_sec)
            conn.execute(
                """
                UPDATE file_enrich_queue
                SET status = :status,
                    attempt_count = :attempt_count,
                    last_error = :last_error,
                    next_retry_at = :next_retry_at,
                    updated_at = :updated_at
                WHERE job_id = :job_id
                """,
                {
                    "job_id": job_id,
                    "status": status,
                    "attempt_count": attempt_count,
                    "last_error": error_message,
                    "next_retry_at": next_retry_at,
                    "updated_at": next_retry_at,
                },
            )
            conn.commit()

    def mark_failed_with_backoff_many(self, updates: list[FileEnrichFailureUpdateDTO]) -> None:
        """보강 작업 실패 상태를 배치 백오프로 갱신한다."""
        if len(updates) == 0:
            return
        with connect(self._db_path) as conn:
            for update in updates:
                row = conn.execute(
                    "SELECT attempt_count FROM file_enrich_queue WHERE job_id = :job_id",
                    {"job_id": update.job_id},
                ).fetchone()
                if row is None:
                    continue
                attempt_count = row_int(row, "attempt_count") + 1
                status = "DEAD" if attempt_count >= update.dead_threshold else "FAILED"
                next_retry_at = self._compute_next_retry_at(
                    now_iso=update.now_iso,
                    attempt_count=attempt_count,
                    backoff_base_sec=update.backoff_base_sec,
                )
                conn.execute(
                    """
                    UPDATE file_enrich_queue
                    SET status = :status,
                        attempt_count = :attempt_count,
                        last_error = :last_error,
                        next_retry_at = :next_retry_at,
                        updated_at = :updated_at
                    WHERE job_id = :job_id
                    """,
                    {
                        "job_id": update.job_id,
                        "status": status,
                        "attempt_count": attempt_count,
                        "last_error": update.error_message,
                        "next_retry_at": next_retry_at,
                        "updated_at": next_retry_at,
                    },
                )
            conn.commit()

    def get_status_counts(self) -> dict[str, int]:
        """큐 상태별 개수를 반환한다."""
        with connect(self._db_path) as conn:
            rows = conn.execute(
                """
                SELECT status, COUNT(*) AS cnt
                FROM file_enrich_queue
                GROUP BY status
                """
            ).fetchall()
        counts: dict[str, int] = {"PENDING": 0, "RUNNING": 0, "FAILED": 0, "DONE": 0, "DEAD": 0}
        for row in rows:
            status = row_str(row, "status")
            counts[status] = row_int(row, "cnt")
        return counts

    def reset_running_to_failed(self, now_iso: str) -> int:
        """비정상 종료 대비 RUNNING 상태를 FAILED로 복구한다."""
        with connect(self._db_path) as conn:
            cur = conn.execute(
                """
                UPDATE file_enrich_queue
                SET status = 'FAILED',
                    last_error = COALESCE(last_error, 'worker interrupted'),
                    next_retry_at = :now_iso,
                    updated_at = :now_iso
                WHERE status = 'RUNNING'
                """,
                {"now_iso": now_iso},
            )
            conn.commit()
            return int(cur.rowcount if cur.rowcount is not None else 0)

    def list_dead(self, repo_root: str, limit: int) -> list[DeadJobItemDTO]:
        """저장소 기준 DEAD 작업 목록을 조회한다."""
        with connect(self._db_path) as conn:
            rows = conn.execute(
                """
                SELECT job_id, repo_root, relative_path, attempt_count, last_error, updated_at
                FROM file_enrich_queue
                WHERE repo_root = :repo_root
                  AND status = 'DEAD'
                ORDER BY updated_at ASC
                LIMIT :limit
                """,
                {"repo_root": repo_root, "limit": limit},
            ).fetchall()
        return [
            DeadJobItemDTO(
                job_id=row_str(row, "job_id"),
                repo_root=row_str(row, "repo_root"),
                relative_path=row_str(row, "relative_path"),
                attempt_count=row_int(row, "attempt_count"),
                last_error=row_optional_str(row, "last_error"),
                updated_at=row_str(row, "updated_at"),
            )
            for row in rows
        ]

    def requeue_dead(self, repo_root: str, limit: int, now_iso: str) -> int:
        """저장소 기준 DEAD 작업을 FAILED로 재큐잉한다."""
        with connect(self._db_path) as conn:
            cur = conn.execute(
                """
                UPDATE file_enrich_queue
                SET status = 'FAILED',
                    next_retry_at = :now_iso,
                    updated_at = :now_iso,
                    last_error = COALESCE(last_error, 'requeued by operator')
                WHERE job_id IN (
                    SELECT job_id
                    FROM file_enrich_queue
                    WHERE repo_root = :repo_root
                      AND status = 'DEAD'
                    ORDER BY updated_at ASC
                    LIMIT :limit
                )
                """,
                {"repo_root": repo_root, "limit": limit, "now_iso": now_iso},
            )
            conn.commit()
            return int(cur.rowcount if cur.rowcount is not None else 0)

    def requeue_dead_all(self, now_iso: str) -> int:
        """저장소 조건 없이 DEAD 작업 전체를 FAILED로 재큐잉한다."""
        with connect(self._db_path) as conn:
            cur = conn.execute(
                """
                UPDATE file_enrich_queue
                SET status = 'FAILED',
                    next_retry_at = :now_iso,
                    updated_at = :now_iso,
                    last_error = COALESCE(last_error, 'requeued by operator')
                WHERE status = 'DEAD'
                """,
                {"now_iso": now_iso},
            )
            conn.commit()
            return int(cur.rowcount if cur.rowcount is not None else 0)

    def purge_dead(self, repo_root: str, limit: int) -> int:
        """저장소 기준 DEAD 작업을 삭제한다."""
        with connect(self._db_path) as conn:
            cur = conn.execute(
                """
                DELETE FROM file_enrich_queue
                WHERE job_id IN (
                    SELECT job_id
                    FROM file_enrich_queue
                    WHERE repo_root = :repo_root
                      AND status = 'DEAD'
                    ORDER BY updated_at ASC
                    LIMIT :limit
                )
                """,
                {"repo_root": repo_root, "limit": limit},
            )
            conn.commit()
            return int(cur.rowcount if cur.rowcount is not None else 0)

    def purge_dead_all(self) -> int:
        """저장소 조건 없이 DEAD 작업 전체를 삭제한다."""
        with connect(self._db_path) as conn:
            cur = conn.execute("DELETE FROM file_enrich_queue WHERE status = 'DEAD'")
            conn.commit()
            return int(cur.rowcount if cur.rowcount is not None else 0)

    def get_status(self, job_id: str) -> str | None:
        """작업의 현재 상태를 조회한다."""
        with connect(self._db_path) as conn:
            row = conn.execute(
                "SELECT status FROM file_enrich_queue WHERE job_id = :job_id",
                {"job_id": job_id},
            ).fetchone()
        if row is None:
            return None
        return row_str(row, "status")

    def _compute_next_retry_at(self, now_iso: str, attempt_count: int, backoff_base_sec: int) -> str:
        """재시도 시각을 지수 백오프로 계산한다."""
        try:
            now_dt = datetime.fromisoformat(now_iso)
        except ValueError:
            now_dt = datetime.utcnow()
        delay_sec = backoff_base_sec * (2 ** max(0, attempt_count - 1))
        return (now_dt + timedelta(seconds=delay_sec)).isoformat()



"""워크스페이스 저장소를 구현한다."""

from pathlib import Path

from sari.core.exceptions import ErrorContext, ValidationError
from sari.core.models import WorkspaceDTO
from sari.db.row_mapper import row_bool, row_optional_str, row_str
from sari.db.schema import connect


class WorkspaceRepository:
    """워크스페이스 영속화를 담당한다."""

    def __init__(self, db_path: Path) -> None:
        """저장소에 사용할 DB 경로를 저장한다."""
        self._db_path = db_path

    def add(self, workspace: WorkspaceDTO) -> None:
        """워크스페이스를 추가한다."""
        with connect(self._db_path) as conn:
            conn.execute(
                """
                INSERT INTO workspaces(path, name, indexed_at, is_active)
                VALUES(:path, :name, :indexed_at, :is_active)
                """,
                workspace.to_sql_params(),
            )
            conn.commit()

    def get_by_path(self, path: str) -> WorkspaceDTO | None:
        """경로로 워크스페이스를 조회한다."""
        with connect(self._db_path) as conn:
            row = conn.execute(
                """
                SELECT path, name, indexed_at, is_active
                FROM workspaces
                WHERE path = :path
                """,
                {"path": path},
            ).fetchone()
        if row is None:
            return None
        return self._workspace_from_row(row)

    def list_all(self) -> list[WorkspaceDTO]:
        """전체 워크스페이스를 조회한다."""
        with connect(self._db_path) as conn:
            rows = conn.execute(
                """
                SELECT path, name, indexed_at, is_active
                FROM workspaces
                ORDER BY path ASC
                """
            ).fetchall()
        results: list[WorkspaceDTO] = []
        for row in rows:
            results.append(self._workspace_from_row(row))
        return results

    def remove(self, path: str) -> None:
        """워크스페이스를 삭제한다."""
        with connect(self._db_path) as conn:
            conn.execute("DELETE FROM workspaces WHERE path = :path", {"path": path})
            conn.commit()

    def _workspace_from_row(self, row: object) -> WorkspaceDTO:
        """DB Row를 WorkspaceDTO로 엄격 매핑한다."""
        if not hasattr(row, "__getitem__"):
            raise ValidationError(ErrorContext(code="ERR_DB_MAPPING_INVALID", message="workspace row 형식이 올바르지 않습니다"))
        return WorkspaceDTO(
            path=row_str(row, "path"),  # type: ignore[arg-type]
            name=row_optional_str(row, "name"),  # type: ignore[arg-type]
            indexed_at=row_optional_str(row, "indexed_at"),  # type: ignore[arg-type]
            is_active=row_bool(row, "is_active"),  # type: ignore[arg-type]
        )



"""LSP 언어/확장자 SSOT 레지스트리를 제공한다."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from solidlsp.ls_config import Language


@dataclass(frozen=True)
class LanguageSupportEntry:
    """운영에서 활성화할 언어와 확장자 집합을 표현한다."""

    language: Language
    extensions: tuple[str, ...]


_LANGUAGE_SUPPORT_ENTRIES: tuple[LanguageSupportEntry, ...] = (
    LanguageSupportEntry(language=Language.PYTHON, extensions=(".py", ".pyi")),
    LanguageSupportEntry(language=Language.TYPESCRIPT, extensions=(".ts", ".tsx", ".js", ".jsx", ".mts", ".cts", ".mjs", ".cjs")),
    LanguageSupportEntry(language=Language.JAVA, extensions=(".java",)),
    LanguageSupportEntry(language=Language.KOTLIN, extensions=(".kt", ".kts")),
    LanguageSupportEntry(language=Language.GO, extensions=(".go",)),
    LanguageSupportEntry(language=Language.RUST, extensions=(".rs",)),
    LanguageSupportEntry(language=Language.CSHARP, extensions=(".cs",)),
    LanguageSupportEntry(language=Language.RUBY, extensions=(".rb", ".erb")),
    LanguageSupportEntry(language=Language.DART, extensions=(".dart",)),
    LanguageSupportEntry(language=Language.CPP, extensions=(".cpp", ".cc", ".cxx", ".c", ".hpp", ".hxx", ".h")),
    LanguageSupportEntry(language=Language.PHP, extensions=(".php",)),
    LanguageSupportEntry(language=Language.R, extensions=(".r", ".rmd", ".rnw")),
    LanguageSupportEntry(language=Language.PERL, extensions=(".pl", ".pm", ".t")),
    LanguageSupportEntry(language=Language.CLOJURE, extensions=(".clj", ".cljs", ".cljc", ".edn")),
    LanguageSupportEntry(language=Language.ELIXIR, extensions=(".ex", ".exs")),
    LanguageSupportEntry(language=Language.ELM, extensions=(".elm",)),
    LanguageSupportEntry(language=Language.TERRAFORM, extensions=(".tf", ".tfvars", ".tfstate")),
    LanguageSupportEntry(language=Language.SWIFT, extensions=(".swift",)),
    LanguageSupportEntry(language=Language.BASH, extensions=(".sh", ".bash")),
    LanguageSupportEntry(language=Language.ZIG, extensions=(".zig", ".zon")),
    LanguageSupportEntry(language=Language.LUA, extensions=(".lua",)),
    LanguageSupportEntry(language=Language.NIX, extensions=(".nix",)),
    LanguageSupportEntry(language=Language.ERLANG, extensions=(".erl", ".hrl", ".escript")),
    LanguageSupportEntry(language=Language.AL, extensions=(".al", ".dal")),
    LanguageSupportEntry(language=Language.FSHARP, extensions=(".fs", ".fsx", ".fsi")),
    LanguageSupportEntry(language=Language.REGO, extensions=(".rego",)),
    LanguageSupportEntry(language=Language.SCALA, extensions=(".scala", ".sbt")),
    LanguageSupportEntry(language=Language.JULIA, extensions=(".jl",)),
    LanguageSupportEntry(language=Language.FORTRAN, extensions=(".f90", ".f95", ".f03", ".f08", ".f", ".for", ".fpp")),
    LanguageSupportEntry(language=Language.HASKELL, extensions=(".hs", ".lhs")),
    LanguageSupportEntry(language=Language.GROOVY, extensions=(".groovy", ".gvy")),
    LanguageSupportEntry(language=Language.VUE, extensions=(".vue",)),
    LanguageSupportEntry(language=Language.POWERSHELL, extensions=(".ps1", ".psm1", ".psd1")),
    LanguageSupportEntry(language=Language.PASCAL, extensions=(".pas", ".pp", ".lpr", ".dpr", ".dpk", ".inc")),
    LanguageSupportEntry(language=Language.MATLAB, extensions=(".m", ".mlx", ".mlapp")),
    LanguageSupportEntry(language=Language.MARKDOWN, extensions=(".md", ".markdown")),
    LanguageSupportEntry(language=Language.YAML, extensions=(".yaml", ".yml")),
    LanguageSupportEntry(language=Language.TOML, extensions=(".toml",)),
)

_CRITICAL_LANGUAGE_NAMES: tuple[str, ...] = (
    "python",
    "typescript",
    "java",
    "kotlin",
    "go",
    "rust",
    "csharp",
)


def iter_language_support_entries() -> tuple[LanguageSupportEntry, ...]:
    """활성 언어 엔트리를 반환한다."""
    return _LANGUAGE_SUPPORT_ENTRIES


def get_enabled_languages() -> tuple[Language, ...]:
    """운영 활성 언어 목록을 반환한다."""
    return tuple(entry.language for entry in _LANGUAGE_SUPPORT_ENTRIES)


def get_enabled_language_names() -> tuple[str, ...]:
    """운영 활성 언어 이름 목록을 반환한다."""
    return tuple(entry.language.value for entry in _LANGUAGE_SUPPORT_ENTRIES)


def get_critical_language_names() -> tuple[str, ...]:
    """하드 게이트에서 100% 성공이 필요한 핵심 언어 목록을 반환한다."""
    enabled = set(get_enabled_language_names())
    return tuple(name for name in _CRITICAL_LANGUAGE_NAMES if name in enabled)


def normalize_language_filter(raw_values: list[str] | tuple[str, ...] | None) -> tuple[str, ...] | None:
    """입력 언어 필터를 정규화한다."""
    if raw_values is None:
        return None
    normalized: list[str] = []
    for raw in raw_values:
        name = str(raw).strip().lower()
        if name == "":
            continue
        normalized.append(name)
    if len(normalized) == 0:
        return None
    enabled = set(get_enabled_language_names())
    invalid = [name for name in normalized if name not in enabled]
    if len(invalid) > 0:
        raise ValueError(f"unsupported language filter: {', '.join(sorted(set(invalid)))}")
    deduped = sorted(set(normalized))
    return tuple(deduped)


def get_default_collection_extensions() -> tuple[str, ...]:
    """수집 기본 확장자 목록을 반환한다."""
    extension_set: set[str] = set()
    for entry in _LANGUAGE_SUPPORT_ENTRIES:
        extension_set.update(entry.extensions)
    return tuple(sorted(extension_set))


def resolve_language_from_path(file_path: str) -> Language | None:
    """파일 경로 기반으로 활성 언어를 결정한다."""
    suffix = Path(file_path).suffix.lower()
    if suffix == "":
        return None
    for entry in _LANGUAGE_SUPPORT_ENTRIES:
        if suffix in entry.extensions:
            return entry.language
    return None



"""SQLite 스키마를 관리한다."""

from pathlib import Path
import sqlite3

from sari.db.migration import ensure_migrated


SCHEMA_SQL = """
CREATE TABLE IF NOT EXISTS workspaces (
    path TEXT PRIMARY KEY,
    name TEXT NULL,
    indexed_at TEXT NULL,
    is_active INTEGER NOT NULL CHECK (is_active IN (0, 1))
);

CREATE TABLE IF NOT EXISTS daemon_runtime (
    singleton_key TEXT PRIMARY KEY,
    pid INTEGER NOT NULL,
    host TEXT NOT NULL,
    port INTEGER NOT NULL,
    state TEXT NOT NULL,
    started_at TEXT NOT NULL,
    session_count INTEGER NOT NULL,
    last_heartbeat_at TEXT NOT NULL,
    last_exit_reason TEXT NULL
);

CREATE TABLE IF NOT EXISTS daemon_runtime_history (
    event_id INTEGER PRIMARY KEY AUTOINCREMENT,
    pid INTEGER NOT NULL,
    exit_reason TEXT NOT NULL,
    occurred_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_daemon_runtime_history_time
ON daemon_runtime_history(occurred_at DESC);

CREATE TABLE IF NOT EXISTS daemon_registry (
    daemon_id TEXT PRIMARY KEY,
    host TEXT NOT NULL,
    port INTEGER NOT NULL,
    pid INTEGER NOT NULL,
    workspace_root TEXT NOT NULL,
    protocol TEXT NOT NULL,
    started_at TEXT NOT NULL,
    last_seen_at TEXT NOT NULL,
    is_draining INTEGER NOT NULL CHECK (is_draining IN (0, 1)),
    deployment_state TEXT NOT NULL DEFAULT 'ACTIVE',
    health_fail_streak INTEGER NOT NULL DEFAULT 0,
    last_health_error TEXT NULL,
    last_health_at TEXT NULL
);

CREATE INDEX IF NOT EXISTS idx_daemon_registry_workspace
ON daemon_registry(workspace_root, is_draining, deployment_state, last_seen_at DESC);

CREATE INDEX IF NOT EXISTS idx_daemon_registry_seen
ON daemon_registry(last_seen_at DESC);

CREATE TABLE IF NOT EXISTS lsp_symbol_cache (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    query TEXT NOT NULL,
    file_hash TEXT NOT NULL,
    items_json TEXT NOT NULL,
    invalidated INTEGER NOT NULL DEFAULT 0 CHECK (invalidated IN (0, 1)),
    updated_at TEXT NOT NULL,
    PRIMARY KEY(repo_root, relative_path, query)
);

CREATE INDEX IF NOT EXISTS idx_lsp_symbol_cache_lookup
ON lsp_symbol_cache(repo_root, query, invalidated);

CREATE TABLE IF NOT EXISTS collected_files_l1 (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    absolute_path TEXT NOT NULL,
    repo_label TEXT NOT NULL,
    mtime_ns INTEGER NOT NULL,
    size_bytes INTEGER NOT NULL,
    content_hash TEXT NOT NULL,
    is_deleted INTEGER NOT NULL DEFAULT 0 CHECK (is_deleted IN (0, 1)),
    last_seen_at TEXT NOT NULL,
    updated_at TEXT NOT NULL,
    enrich_state TEXT NOT NULL,
    PRIMARY KEY(repo_root, relative_path)
);

CREATE INDEX IF NOT EXISTS idx_collected_files_l1_repo
ON collected_files_l1(repo_root, is_deleted);

CREATE INDEX IF NOT EXISTS idx_collected_files_l1_label
ON collected_files_l1(repo_label, is_deleted);

CREATE INDEX IF NOT EXISTS idx_collected_files_l1_seen
ON collected_files_l1(repo_root, last_seen_at);

CREATE TABLE IF NOT EXISTS file_enrich_queue (
    job_id TEXT PRIMARY KEY,
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    content_hash TEXT NOT NULL,
    content_raw TEXT NOT NULL DEFAULT '',
    content_encoding TEXT NOT NULL DEFAULT 'utf-8',
    priority INTEGER NOT NULL DEFAULT 30,
    enqueue_source TEXT NOT NULL DEFAULT 'scan',
    status TEXT NOT NULL,
    attempt_count INTEGER NOT NULL,
    last_error TEXT NULL,
    next_retry_at TEXT NOT NULL,
    created_at TEXT NOT NULL,
    updated_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_file_enrich_queue_status
ON file_enrich_queue(status, next_retry_at);

CREATE INDEX IF NOT EXISTS idx_file_enrich_queue_path
ON file_enrich_queue(repo_root, relative_path);

CREATE INDEX IF NOT EXISTS idx_file_enrich_queue_sched
ON file_enrich_queue(status, priority DESC, next_retry_at, created_at);

CREATE TABLE IF NOT EXISTS candidate_index_changes (
    change_id INTEGER PRIMARY KEY AUTOINCREMENT,
    change_type TEXT NOT NULL,
    status TEXT NOT NULL,
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    absolute_path TEXT NULL,
    content_hash TEXT NULL,
    mtime_ns INTEGER NULL,
    size_bytes INTEGER NULL,
    event_source TEXT NOT NULL,
    reason TEXT NULL,
    created_at TEXT NOT NULL,
    updated_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_candidate_index_changes_status
ON candidate_index_changes(status, change_id);

CREATE INDEX IF NOT EXISTS idx_candidate_index_changes_path_status
ON candidate_index_changes(repo_root, relative_path, status);

CREATE TABLE IF NOT EXISTS collected_file_bodies_l2 (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    content_hash TEXT NOT NULL,
    content_zlib BLOB NOT NULL,
    content_len INTEGER NOT NULL,
    normalized_text TEXT NOT NULL,
    created_at TEXT NOT NULL,
    updated_at TEXT NOT NULL,
    PRIMARY KEY(repo_root, relative_path, content_hash)
);

CREATE TABLE IF NOT EXISTS lsp_symbols (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    content_hash TEXT NOT NULL,
    name TEXT NOT NULL,
    kind TEXT NOT NULL,
    line INTEGER NOT NULL,
    end_line INTEGER NOT NULL,
    symbol_key TEXT NULL,
    parent_symbol_key TEXT NULL,
    depth INTEGER NOT NULL DEFAULT 0,
    container_name TEXT NULL,
    created_at TEXT NOT NULL,
    PRIMARY KEY(repo_root, relative_path, content_hash, name, kind, line, end_line)
);

CREATE INDEX IF NOT EXISTS idx_lsp_symbols_repo_path_depth
ON lsp_symbols(repo_root, relative_path, depth, line);

CREATE INDEX IF NOT EXISTS idx_lsp_symbols_symbol_key
ON lsp_symbols(repo_root, relative_path, content_hash, symbol_key);

CREATE TABLE IF NOT EXISTS lsp_call_relations (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    content_hash TEXT NOT NULL,
    from_symbol TEXT NOT NULL,
    to_symbol TEXT NOT NULL,
    line INTEGER NOT NULL,
    created_at TEXT NOT NULL,
    PRIMARY KEY(repo_root, relative_path, content_hash, from_symbol, to_symbol, line)
);

CREATE TABLE IF NOT EXISTS symbol_importance_cache (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    symbol_name TEXT NOT NULL,
    reference_count INTEGER NOT NULL,
    revision_epoch INTEGER NOT NULL DEFAULT 0,
    updated_at TEXT NOT NULL,
    PRIMARY KEY(repo_root, symbol_name)
);

CREATE INDEX IF NOT EXISTS idx_symbol_importance_repo_count
ON symbol_importance_cache(repo_root, reference_count DESC);

CREATE TABLE IF NOT EXISTS tool_readiness_state (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    content_hash TEXT NOT NULL,
    list_files_ready INTEGER NOT NULL CHECK (list_files_ready IN (0, 1)),
    read_file_ready INTEGER NOT NULL CHECK (read_file_ready IN (0, 1)),
    search_symbol_ready INTEGER NOT NULL CHECK (search_symbol_ready IN (0, 1)),
    get_callers_ready INTEGER NOT NULL CHECK (get_callers_ready IN (0, 1)),
    consistency_ready INTEGER NOT NULL CHECK (consistency_ready IN (0, 1)),
    quality_ready INTEGER NOT NULL CHECK (quality_ready IN (0, 1)),
    tool_ready INTEGER NOT NULL CHECK (tool_ready IN (0, 1)),
    last_reason TEXT NOT NULL,
    updated_at TEXT NOT NULL,
    PRIMARY KEY(repo_root, relative_path)
);

CREATE TABLE IF NOT EXISTS pipeline_policy (
    singleton_key TEXT PRIMARY KEY,
    deletion_hold INTEGER NOT NULL CHECK (deletion_hold IN (0, 1)),
    l3_p95_threshold_ms INTEGER NOT NULL,
    dead_ratio_threshold_bps INTEGER NOT NULL,
    enrich_worker_count INTEGER NOT NULL,
    bootstrap_mode_enabled INTEGER NOT NULL DEFAULT 0 CHECK (bootstrap_mode_enabled IN (0, 1)),
    bootstrap_l3_worker_count INTEGER NOT NULL DEFAULT 1,
    bootstrap_l3_queue_max INTEGER NOT NULL DEFAULT 1000,
    bootstrap_exit_min_l2_coverage_bps INTEGER NOT NULL DEFAULT 9500,
    bootstrap_exit_max_sec INTEGER NOT NULL DEFAULT 1800,
    alert_window_sec INTEGER NOT NULL,
    updated_at TEXT NOT NULL
);

CREATE TABLE IF NOT EXISTS pipeline_control_state (
    singleton_key TEXT PRIMARY KEY,
    auto_hold_enabled INTEGER NOT NULL CHECK (auto_hold_enabled IN (0, 1)),
    auto_hold_active INTEGER NOT NULL CHECK (auto_hold_active IN (0, 1)),
    last_action TEXT NOT NULL,
    updated_at TEXT NOT NULL
);

CREATE TABLE IF NOT EXISTS pipeline_job_events (
    event_id TEXT PRIMARY KEY,
    job_id TEXT NOT NULL,
    status TEXT NOT NULL,
    latency_ms INTEGER NOT NULL,
    created_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_pipeline_job_events_time
ON pipeline_job_events(created_at);

CREATE INDEX IF NOT EXISTS idx_pipeline_job_events_status
ON pipeline_job_events(created_at, status);

CREATE TABLE IF NOT EXISTS pipeline_error_events (
    event_id TEXT PRIMARY KEY,
    occurred_at TEXT NOT NULL,
    component TEXT NOT NULL,
    phase TEXT NOT NULL,
    severity TEXT NOT NULL,
    scope_type TEXT NOT NULL DEFAULT 'GLOBAL' CHECK (scope_type IN ('GLOBAL', 'REPO')),
    repo_root TEXT NULL,
    relative_path TEXT NULL,
    job_id TEXT NULL,
    attempt_count INTEGER NOT NULL,
    error_code TEXT NOT NULL,
    error_message TEXT NOT NULL,
    error_type TEXT NOT NULL,
    stacktrace_text TEXT NOT NULL,
    context_json TEXT NOT NULL,
    worker_name TEXT NOT NULL,
    run_mode TEXT NOT NULL,
    resolved INTEGER NOT NULL CHECK (resolved IN (0, 1)),
    resolved_at TEXT NULL
);

CREATE INDEX IF NOT EXISTS idx_pipeline_error_events_time
ON pipeline_error_events(occurred_at DESC);

CREATE INDEX IF NOT EXISTS idx_pipeline_error_events_component
ON pipeline_error_events(component, phase);

CREATE INDEX IF NOT EXISTS idx_pipeline_error_events_repo_path
ON pipeline_error_events(repo_root, relative_path);

CREATE INDEX IF NOT EXISTS idx_pipeline_error_events_code
ON pipeline_error_events(error_code);

CREATE INDEX IF NOT EXISTS idx_pipeline_error_events_scope_time
ON pipeline_error_events(scope_type, occurred_at DESC);

CREATE TABLE IF NOT EXISTS pipeline_benchmark_runs (
    run_id TEXT PRIMARY KEY,
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    target_files INTEGER NOT NULL,
    profile TEXT NOT NULL,
    started_at TEXT NOT NULL,
    finished_at TEXT NULL,
    status TEXT NOT NULL,
    summary_json TEXT NULL
);

CREATE INDEX IF NOT EXISTS idx_pipeline_benchmark_runs_started
ON pipeline_benchmark_runs(started_at DESC);

CREATE TABLE IF NOT EXISTS pipeline_quality_runs (
    run_id TEXT PRIMARY KEY,
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    limit_files INTEGER NOT NULL,
    profile TEXT NOT NULL,
    started_at TEXT NOT NULL,
    finished_at TEXT NULL,
    status TEXT NOT NULL,
    summary_json TEXT NULL
);

CREATE INDEX IF NOT EXISTS idx_pipeline_quality_runs_started
ON pipeline_quality_runs(started_at DESC);

CREATE TABLE IF NOT EXISTS pipeline_lsp_matrix_runs (
    run_id TEXT PRIMARY KEY,
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    required_languages_json TEXT NOT NULL,
    fail_on_unavailable INTEGER NOT NULL CHECK (fail_on_unavailable IN (0, 1)),
    strict_symbol_gate INTEGER NOT NULL CHECK (strict_symbol_gate IN (0, 1)) DEFAULT 1,
    started_at TEXT NOT NULL,
    finished_at TEXT NULL,
    status TEXT NOT NULL,
    summary_json TEXT NULL
);

CREATE INDEX IF NOT EXISTS idx_pipeline_lsp_matrix_runs_started
ON pipeline_lsp_matrix_runs(started_at DESC);

CREATE TABLE IF NOT EXISTS language_probe_status (
    language TEXT PRIMARY KEY,
    enabled INTEGER NOT NULL CHECK (enabled IN (0, 1)),
    available INTEGER NOT NULL CHECK (available IN (0, 1)),
    last_probe_at TEXT NULL,
    last_error_code TEXT NULL,
    last_error_message TEXT NULL,
    updated_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_language_probe_status_available
ON language_probe_status(available, updated_at DESC);

CREATE TABLE IF NOT EXISTS file_embeddings (
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    relative_path TEXT NOT NULL,
    content_hash TEXT NOT NULL,
    model_id TEXT NOT NULL,
    dim INTEGER NOT NULL,
    vector_json TEXT NOT NULL,
    updated_at TEXT NOT NULL,
    PRIMARY KEY(repo_root, relative_path, content_hash, model_id)
);

CREATE INDEX IF NOT EXISTS idx_file_embeddings_lookup
ON file_embeddings(repo_root, relative_path, model_id);

CREATE TABLE IF NOT EXISTS query_embeddings (
    query_hash TEXT NOT NULL,
    model_id TEXT NOT NULL,
    dim INTEGER NOT NULL,
    vector_json TEXT NOT NULL,
    updated_at TEXT NOT NULL,
    PRIMARY KEY(query_hash, model_id)
);

CREATE TABLE IF NOT EXISTS snippet_entries (
    snippet_id INTEGER PRIMARY KEY AUTOINCREMENT,
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    source_path TEXT NOT NULL,
    start_line INTEGER NOT NULL,
    end_line INTEGER NOT NULL,
    tag TEXT NOT NULL,
    note TEXT NULL,
    commit_hash TEXT NULL,
    content_text TEXT NOT NULL,
    created_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_snippet_entries_repo_tag
ON snippet_entries(repo_root, tag, created_at DESC);

CREATE INDEX IF NOT EXISTS idx_snippet_entries_repo_path
ON snippet_entries(repo_root, source_path, created_at DESC);

CREATE TABLE IF NOT EXISTS knowledge_entries (
    entry_id INTEGER PRIMARY KEY AUTOINCREMENT,
    kind TEXT NOT NULL,
    repo_root TEXT NOT NULL CHECK (repo_root <> ''),
    topic TEXT NOT NULL,
    content_text TEXT NOT NULL,
    tags_json TEXT NOT NULL,
    related_files_json TEXT NOT NULL,
    created_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_knowledge_entries_repo_kind
ON knowledge_entries(repo_root, kind, created_at DESC);

CREATE INDEX IF NOT EXISTS idx_knowledge_entries_repo_topic
ON knowledge_entries(repo_root, topic, created_at DESC);
"""

def ensure_parent_dir(db_path: Path) -> None:
    """DB 상위 디렉터리를 보장한다."""
    db_path.parent.mkdir(parents=True, exist_ok=True)


def connect(db_path: Path) -> sqlite3.Connection:
    """SQLite 연결을 생성하고 로우 팩토리를 설정한다."""
    conn = sqlite3.connect(str(db_path))
    # busy timeout/synchronous는 연결 단위 정책으로 항상 적용한다.
    conn.execute("PRAGMA synchronous=NORMAL")
    conn.execute("PRAGMA busy_timeout=5000")
    conn.row_factory = sqlite3.Row
    return conn


def init_schema(db_path: Path) -> None:
    """필수 테이블을 생성한다."""
    ensure_parent_dir(db_path)
    with connect(db_path) as conn:
        # WAL은 초기화(쓰기 가능) 단계에서만 고정한다.
        conn.execute("PRAGMA journal_mode=WAL")
        conn.executescript(SCHEMA_SQL)
        conn.commit()
    # 스키마 마이그레이션의 단일 진실 소스는 Alembic head다.
    ensure_migrated(db_path)



"""
Provides Markdown specific instantiation of the LanguageServer class using marksman.
Contains various configurations and settings specific to Markdown.
"""

import logging
import os
import pathlib

from overrides import override

from solidlsp.ls import SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

from .common import RuntimeDependency, RuntimeDependencyCollection

log = logging.getLogger(__name__)


class Marksman(SolidLanguageServer):
    """
    Provides Markdown specific instantiation of the LanguageServer class using marksman.
    """

    marksman_releases = "https://github.com/artempyanykh/marksman/releases/download/2024-12-18"
    runtime_dependencies = RuntimeDependencyCollection(
        [
            RuntimeDependency(
                id="marksman",
                url=f"{marksman_releases}/marksman-linux-x64",
                platform_id="linux-x64",
                archive_type="binary",
                binary_name="marksman",
            ),
            RuntimeDependency(
                id="marksman",
                url=f"{marksman_releases}/marksman-linux-arm64",
                platform_id="linux-arm64",
                archive_type="binary",
                binary_name="marksman",
            ),
            RuntimeDependency(
                id="marksman",
                url=f"{marksman_releases}/marksman-macos",
                platform_id="osx-x64",
                archive_type="binary",
                binary_name="marksman",
            ),
            RuntimeDependency(
                id="marksman",
                url=f"{marksman_releases}/marksman-macos",
                platform_id="osx-arm64",
                archive_type="binary",
                binary_name="marksman",
            ),
            RuntimeDependency(
                id="marksman",
                url=f"{marksman_releases}/marksman.exe",
                platform_id="win-x64",
                archive_type="binary",
                binary_name="marksman.exe",
            ),
        ]
    )

    @classmethod
    def _setup_runtime_dependencies(cls, config: LanguageServerConfig, solidlsp_settings: SolidLSPSettings) -> str:
        """Setup runtime dependencies for marksman and return the command to start the server."""
        deps = cls.runtime_dependencies
        dependency = deps.get_single_dep_for_current_platform()

        marksman_ls_dir = cls.ls_resources_dir(solidlsp_settings)
        marksman_executable_path = deps.binary_path(marksman_ls_dir)
        if not os.path.exists(marksman_executable_path):
            log.info(
                f"Downloading marksman from {dependency.url} to {marksman_ls_dir}",
            )
            deps.install(marksman_ls_dir)
        if not os.path.exists(marksman_executable_path):
            raise FileNotFoundError(f"Download failed? Could not find marksman executable at {marksman_executable_path}")
        os.chmod(marksman_executable_path, 0o755)
        return marksman_executable_path

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates a Marksman instance. This class is not meant to be instantiated directly.
        Use LanguageServer.create() instead.
        """
        marksman_executable_path = self._setup_runtime_dependencies(config, solidlsp_settings)

        super().__init__(
            config,
            repository_root_path,
            ProcessLaunchInfo(cmd=f"{marksman_executable_path} server", cwd=repository_root_path),
            "markdown",
            solidlsp_settings,
        )

    @override
    def is_ignored_dirname(self, dirname: str) -> bool:
        return super().is_ignored_dirname(dirname) or dirname in ["node_modules", ".obsidian", ".vitepress", ".vuepress"]

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Marksman Language Server.
        """
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        initialize_params: InitializeParams = {  # type: ignore
            "processId": os.getpid(),
            "locale": "en",
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "capabilities": {
                "textDocument": {
                    "synchronization": {"didSave": True, "dynamicRegistration": True},
                    "completion": {"dynamicRegistration": True, "completionItem": {"snippetSupport": True}},
                    "definition": {"dynamicRegistration": True},
                    "references": {"dynamicRegistration": True},
                    "documentSymbol": {
                        "dynamicRegistration": True,
                        "hierarchicalDocumentSymbolSupport": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},  # type: ignore[arg-type]
                    },
                    "hover": {"dynamicRegistration": True, "contentFormat": ["markdown", "plaintext"]},  # type: ignore[list-item]
                    "codeAction": {"dynamicRegistration": True},
                },
                "workspace": {
                    "workspaceFolders": True,
                    "didChangeConfiguration": {"dynamicRegistration": True},
                    "symbol": {"dynamicRegistration": True},
                },
            },
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }
        return initialize_params

    def _start_server(self) -> None:
        """
        Starts the Marksman Language Server and waits for it to be ready.
        """

        def register_capability_handler(_params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        def do_nothing(_params: dict) -> None:
            return

        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)

        log.info("Starting marksman server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to marksman server and awaiting response")
        init_response = self.server.send.initialize(initialize_params)
        log.debug(f"Received initialize response from marksman server: {init_response}")

        # Verify server capabilities
        assert "textDocumentSync" in init_response["capabilities"]
        assert "completionProvider" in init_response["capabilities"]
        assert "definitionProvider" in init_response["capabilities"]

        self.server.notify.initialized({})

        # marksman is typically ready immediately after initialization
        log.info("Marksman server initialization complete")



"""MCP 운영 도구(doctor/rescan/repo_candidates)를 제공한다."""

from __future__ import annotations

from dataclasses import dataclass

from sari.core.exceptions import ValidationError
from sari.core.models import ErrorResponseDTO
from sari.core.repo_resolver import resolve_repo_root
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.mcp.tools.pack1 import Pack1MetaDTO, pack1_error, pack1_success
from sari.services.admin_service import AdminService


def validate_repo_argument(arguments: dict[str, object], workspace_repo: WorkspaceRepository) -> ErrorResponseDTO | None:
    """repo 인자를 검증하고 오류 DTO를 반환한다."""
    repo = arguments.get("repo")
    if not isinstance(repo, str) or repo.strip() == "":
        return ErrorResponseDTO(code="ERR_REPO_REQUIRED", message="repo is required")
    try:
        resolved_repo = resolve_repo_root(
            repo_or_path=repo.strip(),
            workspace_paths=[item.path for item in workspace_repo.list_all()],
        )
    except ValidationError as exc:
        return ErrorResponseDTO(code=exc.context.code, message=exc.context.message)
    if workspace_repo.get_by_path(resolved_repo) is None:
        return ErrorResponseDTO(code="ERR_REPO_NOT_FOUND", message="repo is not registered workspace")
    arguments["repo"] = resolved_repo
    return None


@dataclass(frozen=True)
class DoctorItemDTO:
    """doctor 응답 항목 DTO다."""

    name: str
    passed: bool
    detail: str

    def to_dict(self) -> dict[str, object]:
        """직렬화 가능한 딕셔너리로 변환한다."""
        return {"name": self.name, "passed": self.passed, "detail": self.detail}


class DoctorTool:
    """doctor MCP 도구를 처리한다."""

    def __init__(self, admin_service: AdminService, workspace_repo: WorkspaceRepository) -> None:
        """필요 의존성을 주입한다."""
        self._admin_service = admin_service
        self._workspace_repo = workspace_repo

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """doctor 응답을 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)

        items = [
            DoctorItemDTO(name=check.name, passed=check.passed, detail=check.detail).to_dict()
            for check in self._admin_service.doctor()
        ]
        return pack1_success(
            {
                "items": items,
                "meta": Pack1MetaDTO(
                    candidate_count=len(items),
                    resolved_count=len(items),
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )


class RescanTool:
    """rescan MCP 도구를 처리한다."""

    def __init__(self, admin_service: AdminService, workspace_repo: WorkspaceRepository) -> None:
        """필요 의존성을 주입한다."""
        self._admin_service = admin_service
        self._workspace_repo = workspace_repo

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """캐시 무효화 결과를 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)

        result = self._admin_service.index()
        invalidated_rows = int(result.get("invalidated_cache_rows", 0))
        return pack1_success(
            {
                "items": [],
                "invalidated_cache_rows": invalidated_rows,
                "meta": Pack1MetaDTO(
                    candidate_count=0,
                    resolved_count=invalidated_rows,
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )


class RepoCandidatesTool:
    """repo_candidates MCP 도구를 처리한다."""

    def __init__(self, admin_service: AdminService, workspace_repo: WorkspaceRepository) -> None:
        """필요 의존성을 주입한다."""
        self._admin_service = admin_service
        self._workspace_repo = workspace_repo

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """저장소 후보 목록을 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)

        items = self._admin_service.repo_candidates()
        return pack1_success(
            {
                "items": items,
                "meta": Pack1MetaDTO(
                    candidate_count=len(items),
                    resolved_count=len(items),
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )



"""
Provides TOML specific instantiation of the LanguageServer class using Taplo.
Contains various configurations and settings specific to TOML files.
"""

import gzip
import hashlib
import logging
import os
import platform
import shutil
import socket
import stat
import urllib.error
import urllib.request

# Download timeout in seconds (prevents indefinite hangs)
DOWNLOAD_TIMEOUT_SECONDS = 120

from solidlsp.ls import SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.ls_utils import PathUtils
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)

# Taplo release version and download URLs
TAPLO_VERSION = "0.10.0"
TAPLO_DOWNLOAD_BASE = f"https://github.com/tamasfe/taplo/releases/download/{TAPLO_VERSION}"

# SHA256 checksums for Taplo releases (verified from official GitHub releases)
# Source: https://github.com/tamasfe/taplo/releases/tag/0.10.0
# To update: download each release file and run: sha256sum <filename>
TAPLO_SHA256_CHECKSUMS: dict[str, str] = {
    "taplo-windows-x86_64.zip": "1615eed140039bd58e7089109883b1c434de5d6de8f64a993e6e8c80ca57bdf9",
    "taplo-windows-x86.zip": "b825701daab10dcfc0251e6d668cd1a9c0e351e7f6762dd20844c3f3f3553aa0",
    "taplo-darwin-x86_64.gz": "898122cde3a0b1cd1cbc2d52d3624f23338218c91b5ddb71518236a4c2c10ef2",
    "taplo-darwin-aarch64.gz": "713734314c3e71894b9e77513c5349835eefbd52908445a0d73b0c7dc469347d",
    "taplo-linux-x86_64.gz": "8fe196b894ccf9072f98d4e1013a180306e17d244830b03986ee5e8eabeb6156",
    "taplo-linux-aarch64.gz": "033681d01eec8376c3fd38fa3703c79316f5e14bb013d859943b60a07bccdcc3",
    "taplo-linux-armv7.gz": "6b728896afe2573522f38b8e668b1ff40eb5928fd9d6d0c253ecae508274d417",
}


def _verify_sha256(file_path: str, expected_hash: str) -> bool:
    """Verify SHA256 checksum of a downloaded file."""
    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(8192), b""):
            sha256_hash.update(chunk)
    actual_hash = sha256_hash.hexdigest()
    return actual_hash.lower() == expected_hash.lower()


def _get_taplo_download_url() -> tuple[str, str]:
    """
    Get the appropriate Taplo download URL for the current platform.

    Returns:
        Tuple of (download_url, executable_name)

    """
    system = platform.system().lower()
    machine = platform.machine().lower()

    # Map machine architecture to Taplo naming convention
    arch_map = {
        "x86_64": "x86_64",
        "amd64": "x86_64",
        "x86": "x86",
        "i386": "x86",
        "i686": "x86",
        "aarch64": "aarch64",
        "arm64": "aarch64",
        "armv7l": "armv7",
    }

    arch = arch_map.get(machine, "x86_64")  # Default to x86_64

    if system == "windows":
        filename = f"taplo-windows-{arch}.zip"
        executable = "taplo.exe"
    elif system == "darwin":
        filename = f"taplo-darwin-{arch}.gz"
        executable = "taplo"
    else:  # Linux and others
        filename = f"taplo-linux-{arch}.gz"
        executable = "taplo"

    return f"{TAPLO_DOWNLOAD_BASE}/{filename}", executable


class TaploServer(SolidLanguageServer):
    """
    Provides TOML specific instantiation of the LanguageServer class using Taplo.
    Taplo is a TOML toolkit with LSP support for validation, formatting, and schema support.
    """

    @staticmethod
    def _determine_log_level(line: str) -> int:
        """Classify Taplo stderr output to avoid false-positive errors."""
        line_lower = line.lower()

        # Known informational messages from Taplo
        if any(
            [
                "schema" in line_lower and "not found" in line_lower,
                "warning" in line_lower,
            ]
        ):
            return logging.DEBUG

        return SolidLanguageServer._determine_log_level(line)

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates a TaploServer instance. This class is not meant to be instantiated directly.
        Use LanguageServer.create() instead.
        """
        taplo_executable_path = self._setup_runtime_dependencies(solidlsp_settings)
        super().__init__(
            config,
            repository_root_path,
            ProcessLaunchInfo(cmd=f"{taplo_executable_path} lsp stdio", cwd=repository_root_path),
            "toml",
            solidlsp_settings,
        )

    @classmethod
    def _setup_runtime_dependencies(cls, solidlsp_settings: SolidLSPSettings) -> str:
        """
        Setup runtime dependencies for Taplo and return the command to start the server.
        """
        # First check if taplo is already installed system-wide
        system_taplo = shutil.which("taplo")
        if system_taplo:
            log.info(f"Using system-installed Taplo at: {system_taplo}")
            return system_taplo

        # Setup local installation directory
        taplo_dir = os.path.join(cls.ls_resources_dir(solidlsp_settings), "taplo")
        os.makedirs(taplo_dir, exist_ok=True)

        _, executable_name = _get_taplo_download_url()
        taplo_executable = os.path.join(taplo_dir, executable_name)

        if os.path.exists(taplo_executable) and os.access(taplo_executable, os.X_OK):
            log.info(f"Using cached Taplo at: {taplo_executable}")
            return taplo_executable

        # Download and install Taplo
        log.info(f"Taplo not found. Downloading version {TAPLO_VERSION}...")
        cls._download_taplo(taplo_dir, taplo_executable)

        if not os.path.exists(taplo_executable):
            raise FileNotFoundError(
                f"Taplo executable not found at {taplo_executable}. "
                "Installation may have failed. Try installing manually: cargo install taplo-cli --locked"
            )

        return taplo_executable

    @classmethod
    def _download_taplo(cls, install_dir: str, executable_path: str) -> None:
        """Download and extract Taplo binary with SHA256 verification."""
        # 참고: consider using existing download utilities in SolidLSP instead of the custom logic here
        download_url, _ = _get_taplo_download_url()
        archive_filename = os.path.basename(download_url)

        try:
            log.info(f"Downloading Taplo from: {download_url}")
            archive_path = os.path.join(install_dir, archive_filename)

            # Download the archive with timeout to prevent indefinite hangs
            old_timeout = socket.getdefaulttimeout()
            try:
                socket.setdefaulttimeout(DOWNLOAD_TIMEOUT_SECONDS)
                urllib.request.urlretrieve(download_url, archive_path)
            finally:
                socket.setdefaulttimeout(old_timeout)

            # Verify SHA256 checksum
            expected_hash = TAPLO_SHA256_CHECKSUMS.get(archive_filename)
            if expected_hash:
                if not _verify_sha256(archive_path, expected_hash):
                    os.remove(archive_path)
                    raise RuntimeError(
                        f"SHA256 checksum verification failed for {archive_filename}. "
                        "The downloaded file may be corrupted or tampered with. "
                        "Try installing manually: cargo install taplo-cli --locked"
                    )
                log.info(f"SHA256 checksum verified for {archive_filename}")
            else:
                log.warning(
                    f"No SHA256 checksum available for {archive_filename}. "
                    "Skipping verification - consider installing manually: cargo install taplo-cli --locked"
                )

            # Extract based on format
            if archive_path.endswith(".gz") and not archive_path.endswith(".tar.gz"):
                # Single file gzip
                with gzip.open(archive_path, "rb") as f_in:
                    with open(executable_path, "wb") as f_out:
                        f_out.write(f_in.read())
            elif archive_path.endswith(".zip"):
                import zipfile

                with zipfile.ZipFile(archive_path, "r") as zip_ref:
                    # Security: Validate paths to prevent zip slip vulnerability
                    for member in zip_ref.namelist():
                        member_path = os.path.normpath(os.path.join(install_dir, member))
                        if not member_path.startswith(os.path.normpath(install_dir)):
                            raise RuntimeError(f"Zip slip detected: {member} attempts to escape install directory")
                    zip_ref.extractall(install_dir)

            # Make executable on Unix systems
            if os.name != "nt":
                os.chmod(executable_path, os.stat(executable_path).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

            # Clean up archive
            os.remove(archive_path)
            log.info(f"Taplo installed successfully at: {executable_path}")

        except (OSError, RuntimeError, ValueError, TypeError, urllib.error.URLError, TimeoutError) as e:
            log.error(f"Failed to download Taplo: {e}")
            raise RuntimeError(
                f"Failed to download Taplo from {download_url}. Try installing manually: cargo install taplo-cli --locked"
            ) from e

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Taplo Language Server.
        """
        root_uri = PathUtils.path_to_uri(repository_absolute_path)
        initialize_params = {
            "locale": "en",
            "capabilities": {
                "textDocument": {
                    "synchronization": {"didSave": True, "dynamicRegistration": True},
                    "completion": {"dynamicRegistration": True, "completionItem": {"snippetSupport": True}},
                    "definition": {"dynamicRegistration": True},
                    "references": {"dynamicRegistration": True},
                    "documentSymbol": {
                        "dynamicRegistration": True,
                        "hierarchicalDocumentSymbolSupport": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                    },
                    "hover": {"dynamicRegistration": True, "contentFormat": ["markdown", "plaintext"]},
                    "codeAction": {"dynamicRegistration": True},
                },
                "workspace": {
                    "workspaceFolders": True,
                    "didChangeConfiguration": {"dynamicRegistration": True},
                    "symbol": {"dynamicRegistration": True},
                },
            },
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }
        return initialize_params  # type: ignore

    def _start_server(self) -> None:
        """
        Starts the Taplo Language Server and initializes it.
        """

        def register_capability_handler(params: dict[str, object]) -> None:
            return

        def do_nothing(params: dict[str, object]) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)

        log.info("Starting Taplo server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request to Taplo server")
        init_response = self.server.send.initialize(initialize_params)
        log.debug(f"Received initialize response from Taplo: {init_response}")

        # Verify document symbol support
        capabilities = init_response.get("capabilities", {})
        if capabilities.get("documentSymbolProvider"):
            log.info("Taplo server supports document symbols")
        else:
            log.warning("Taplo server may have limited document symbol support")

        self.server.notify.initialized({})

        log.info("Taplo server initialization complete")

    def is_ignored_dirname(self, dirname: str) -> bool:
        """Define TOML-specific directories to ignore."""
        return super().is_ignored_dirname(dirname) or dirname in ["target", ".cargo", "node_modules"]



"""언어별 LSP readiness probe 서비스를 제공한다."""

from __future__ import annotations

import os
from pathlib import Path
import threading
import uuid
from typing import Callable

from sari.core.exceptions import DaemonError, ErrorContext
from sari.core.language_registry import LanguageSupportEntry, iter_language_support_entries
from sari.core.lsp_provision_policy import get_lsp_provision_policy
from sari.core.models import LanguageProbeStatusDTO, now_iso8601_utc
from sari.db.repositories.language_probe_repository import LanguageProbeRepository
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.lsp.hub import LspHub
from solidlsp.ls_exceptions import SolidLSPException


class LanguageProbeService:
    """레포 단위 언어 readiness를 점검하고 결과를 저장한다."""

    def __init__(
        self,
        workspace_repo: WorkspaceRepository,
        lsp_hub: LspHub,
        probe_repo: LanguageProbeRepository | None = None,
        entries: tuple[LanguageSupportEntry, ...] | None = None,
        now_provider: Callable[[], str] | None = None,
        per_language_timeout_sec: float = 20.0,
    ) -> None:
        """필요 의존성을 저장한다."""
        self._workspace_repo = workspace_repo
        self._lsp_hub = lsp_hub
        self._probe_repo = probe_repo
        self._entries = iter_language_support_entries() if entries is None else entries
        self._now_provider = now_provider if now_provider is not None else now_iso8601_utc
        self._per_language_timeout_sec = max(0.1, float(per_language_timeout_sec))

    def run(self, repo_root: str) -> dict[str, object]:
        """전체 활성 언어에 대한 readiness probe를 실행한다."""
        normalized_repo = str(Path(repo_root).expanduser().resolve())
        self._ensure_registered_repo(normalized_repo)
        started_at = self._now_provider()
        sample_by_extension = self._collect_first_sample_by_extension(normalized_repo)
        items: list[LanguageProbeStatusDTO] = []
        for entry in self._entries:
            item = self._probe_single_language(
                repo_root=normalized_repo,
                entry=entry,
                sample_by_extension=sample_by_extension,
                probe_at=started_at,
            )
            items.append(item)
            if self._probe_repo is not None:
                self._probe_repo.upsert_result(
                    language=item.language,
                    enabled=item.enabled,
                    available=item.available,
                    last_probe_at=item.last_probe_at,
                    last_error_code=item.last_error_code,
                    last_error_message=item.last_error_message,
                )
        total_languages = len(items)
        available_languages = len([item for item in items if item.available])
        return {
            "run_id": str(uuid.uuid4()),
            "repo_root": normalized_repo,
            "started_at": started_at,
            "finished_at": self._now_provider(),
            "summary": {
                "total_languages": total_languages,
                "available_languages": available_languages,
                "unavailable_languages": total_languages - available_languages,
            },
            "languages": [item.to_dict() for item in items],
        }

    def _ensure_registered_repo(self, repo_root: str) -> None:
        """등록되지 않은 repo 입력을 명시적으로 차단한다."""
        workspace = self._workspace_repo.get_by_path(repo_root)
        if workspace is None:
            raise DaemonError(ErrorContext(code="ERR_REPO_NOT_REGISTERED", message=f"repo is not registered: {repo_root}"))

    def _collect_first_sample_by_extension(self, repo_root: str) -> dict[str, str]:
        """레포 내 확장자별 첫 샘플 파일 상대경로를 수집한다."""
        required_extensions: set[str] = set()
        for entry in self._entries:
            for extension in entry.extensions:
                required_extensions.add(extension.lower())
        found: dict[str, str] = {}
        skip_dirs = {".git", "node_modules", "dist", "build", ".venv", "__pycache__"}
        for current_root, dirs, files in os.walk(repo_root):
            dirs[:] = [item for item in dirs if item not in skip_dirs]
            for name in files:
                suffix = Path(name).suffix.lower()
                if suffix == "" or suffix in found or suffix not in required_extensions:
                    continue
                absolute_path = Path(current_root) / name
                relative_path = str(absolute_path.resolve().relative_to(Path(repo_root).resolve()).as_posix())
                found[suffix] = relative_path
            if len(found) == len(required_extensions):
                break
        return found

    def _probe_single_language(
        self,
        repo_root: str,
        entry: LanguageSupportEntry,
        sample_by_extension: dict[str, str],
        probe_at: str,
    ) -> LanguageProbeStatusDTO:
        """단일 언어 readiness를 probe하고 결과 DTO를 반환한다."""
        result_box: list[LanguageProbeStatusDTO] = []
        error_box: list[BaseException] = []
        done = threading.Event()

        def _runner() -> None:
            """단일 언어 probe를 별도 스레드에서 실행한다."""
            try:
                result_box.append(
                    self._probe_single_language_impl(
                        repo_root=repo_root,
                        entry=entry,
                        sample_by_extension=sample_by_extension,
                        probe_at=probe_at,
                    )
                )
            except BaseException as exc:  # pragma: no cover - 예기치 못한 경계 예외
                error_box.append(exc)
            finally:
                done.set()

        worker = threading.Thread(
            target=_runner,
            name=f"sari-language-probe-{entry.language.value}",
            daemon=True,
        )
        worker.start()
        finished = done.wait(timeout=self._per_language_timeout_sec)
        if not finished:
            policy = get_lsp_provision_policy(entry.language.value)
            return LanguageProbeStatusDTO(
                language=entry.language.value,
                enabled=True,
                available=False,
                last_probe_at=probe_at,
                last_error_code="ERR_LSP_TIMEOUT",
                last_error_message=(
                    f"language probe timed out after {self._per_language_timeout_sec:.1f}s: "
                    f"{entry.language.value}"
                ),
                updated_at=probe_at,
                symbol_extract_success=False,
                document_symbol_count=0,
                path_mapping_ok=False,
                timeout_occurred=True,
                recovered_by_restart=False,
                provisioning_mode=policy.provisioning_mode,
                missing_dependency=None,
                install_hint=policy.install_hint,
            )
        if len(error_box) > 0:
            policy = get_lsp_provision_policy(entry.language.value)
            return LanguageProbeStatusDTO(
                language=entry.language.value,
                enabled=True,
                available=False,
                last_probe_at=probe_at,
                last_error_code="ERR_LSP_PROBE_INTERNAL",
                last_error_message=str(error_box[0]),
                updated_at=probe_at,
                symbol_extract_success=False,
                document_symbol_count=0,
                path_mapping_ok=False,
                timeout_occurred=False,
                recovered_by_restart=False,
                provisioning_mode=policy.provisioning_mode,
                missing_dependency=None,
                install_hint=policy.install_hint,
            )
        return result_box[0]

    def _probe_single_language_impl(
        self,
        repo_root: str,
        entry: LanguageSupportEntry,
        sample_by_extension: dict[str, str],
        probe_at: str,
    ) -> LanguageProbeStatusDTO:
        """단일 언어 probe 본체를 수행한다."""
        sample_path = self._pick_sample_path(entry=entry, sample_by_extension=sample_by_extension)
        policy = get_lsp_provision_policy(entry.language.value)
        if sample_path is None:
            return LanguageProbeStatusDTO(
                language=entry.language.value,
                enabled=True,
                available=False,
                last_probe_at=probe_at,
                last_error_code="ERR_LANGUAGE_SAMPLE_NOT_FOUND",
                last_error_message=f"sample file not found for {entry.language.value}",
                updated_at=probe_at,
                symbol_extract_success=False,
                document_symbol_count=0,
                path_mapping_ok=False,
                timeout_occurred=False,
                recovered_by_restart=False,
                provisioning_mode=policy.provisioning_mode,
                missing_dependency=None,
                install_hint=policy.install_hint,
            )
        try:
            lsp = self._lsp_hub.get_or_start(language=entry.language, repo_root=repo_root)
            symbol_items = list(lsp.request_document_symbols(sample_path).iter_symbols())
            return LanguageProbeStatusDTO(
                language=entry.language.value,
                enabled=True,
                available=True,
                last_probe_at=probe_at,
                last_error_code=None,
                last_error_message=None,
                updated_at=probe_at,
                symbol_extract_success=True,
                document_symbol_count=len(symbol_items),
                path_mapping_ok=True,
                timeout_occurred=False,
                recovered_by_restart=False,
                provisioning_mode=policy.provisioning_mode,
                missing_dependency=None,
                install_hint=policy.install_hint,
            )
        except DaemonError as exc:
            classified_code = _classify_lsp_error_code(code=exc.context.code, message=exc.context.message)
            timeout_occurred = _is_timeout_error(code=classified_code, message=exc.context.message)
            return LanguageProbeStatusDTO(
                language=entry.language.value,
                enabled=True,
                available=False,
                last_probe_at=probe_at,
                last_error_code=classified_code,
                last_error_message=exc.context.message,
                updated_at=probe_at,
                symbol_extract_success=False,
                document_symbol_count=0,
                path_mapping_ok=False,
                timeout_occurred=timeout_occurred,
                recovered_by_restart=_is_recovered_by_restart(exc.context.message),
                provisioning_mode=policy.provisioning_mode,
                missing_dependency=_extract_missing_dependency(exc.context.message),
                install_hint=policy.install_hint,
            )
        except SolidLSPException as exc:
            error_message = str(exc)
            code = _extract_error_code(error_message, default_code="ERR_LSP_DOCUMENT_SYMBOL_FAILED")
            classified_code = _classify_lsp_error_code(code=code, message=error_message)
            timeout_occurred = _is_timeout_error(code=classified_code, message=error_message)
            return LanguageProbeStatusDTO(
                language=entry.language.value,
                enabled=True,
                available=False,
                last_probe_at=probe_at,
                last_error_code=classified_code,
                last_error_message=error_message,
                updated_at=probe_at,
                symbol_extract_success=False,
                document_symbol_count=0,
                path_mapping_ok=False,
                timeout_occurred=timeout_occurred,
                recovered_by_restart=_is_recovered_by_restart(error_message),
                provisioning_mode=policy.provisioning_mode,
                missing_dependency=_extract_missing_dependency(error_message),
                install_hint=policy.install_hint,
            )
        except (RuntimeError, OSError, ValueError, TypeError) as exc:
            error_message = str(exc)
            classified_code = _classify_lsp_error_code(code="ERR_LSP_DOCUMENT_SYMBOL_FAILED", message=error_message)
            return LanguageProbeStatusDTO(
                language=entry.language.value,
                enabled=True,
                available=False,
                last_probe_at=probe_at,
                last_error_code=classified_code,
                last_error_message=error_message,
                updated_at=probe_at,
                symbol_extract_success=False,
                document_symbol_count=0,
                path_mapping_ok=False,
                timeout_occurred=_is_timeout_error(code=classified_code, message=error_message),
                recovered_by_restart=_is_recovered_by_restart(error_message),
                provisioning_mode=policy.provisioning_mode,
                missing_dependency=_extract_missing_dependency(error_message),
                install_hint=policy.install_hint,
            )

    def _pick_sample_path(self, entry: LanguageSupportEntry, sample_by_extension: dict[str, str]) -> str | None:
        """언어 엔트리 확장자 목록에서 사용 가능한 첫 샘플 파일을 고른다."""
        for extension in entry.extensions:
            candidate = sample_by_extension.get(extension.lower())
            if candidate is not None:
                return candidate
        return None


def _extract_error_code(message: str, default_code: str) -> str:
    """예외 메시지 선두의 ERR_* 코드를 추출한다."""
    trimmed = message.strip()
    if trimmed.startswith("ERR_"):
        code = trimmed.split(":", 1)[0].strip()
        if code != "":
            return code
    return default_code


def _is_timeout_error(code: str, message: str) -> bool:
    """오류 코드/메시지가 타임아웃 성격인지 판별한다."""
    timeout_codes = {
        "ERR_LSP_TIMEOUT",
        "ERR_LSP_REQUEST_TIMEOUT",
        "ERR_LSP_DOCUMENT_SYMBOL_TIMEOUT",
    }
    normalized_message = message.strip().lower()
    return (code in timeout_codes) or ("timeout" in normalized_message) or ("timed out" in normalized_message)


def _is_recovered_by_restart(message: str) -> bool:
    """메시지에서 재시작 복구 여부 플래그를 탐지한다."""
    normalized_message = message.strip().lower()
    return ("recovered_by_restart" in normalized_message) and ("true" in normalized_message)


def _classify_lsp_error_code(code: str, message: str) -> str:
    """LSP 오류를 정책 코드로 정규화한다."""
    normalized_message = message.strip().lower()
    missing_server_tokens = (
        "command not found",
        "no such file",
        "file not found",
        "not installed",
        "missing executable",
        "failed to spawn",
        "failed to start",
        "cannot find",
        "filenotfounderror",
    )
    if any(token in normalized_message for token in missing_server_tokens):
        return "ERR_LSP_SERVER_MISSING"
    if _is_timeout_error(code=code, message=message):
        return "ERR_LSP_TIMEOUT"
    return code


def _extract_missing_dependency(message: str) -> str | None:
    """예외 메시지에서 누락 의존성 토큰을 추출한다."""
    normalized_message = message.strip()
    if normalized_message == "":
        return None
    lowered = normalized_message.lower()
    if "pyright" in lowered:
        return "pyright"
    if "node" in lowered:
        return "node"
    if "npm" in lowered:
        return "npm"
    if "dotnet" in lowered:
        return "dotnet"
    if "java" in lowered:
        return "java"
    if "no such file" in lowered or "command not found" in lowered or "missing required commands" in lowered:
        return "server_binary"
    return None



import logging
import os
import pathlib
from typing import cast

from solidlsp.ls import SolidLanguageServer
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

from ..ls_config import LanguageServerConfig
from ..lsp_protocol_handler.lsp_types import InitializeParams
from .common import RuntimeDependency, RuntimeDependencyCollection

log = logging.getLogger(__name__)


class DartLanguageServer(SolidLanguageServer):
    """
    Provides Dart specific instantiation of the LanguageServer class. Contains various configurations and settings specific to Dart.
    """

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings) -> None:
        """
        Creates a DartServer instance. This class is not meant to be instantiated directly. Use LanguageServer.create() instead.
        """
        executable_path = self._setup_runtime_dependencies(solidlsp_settings)
        super().__init__(
            config, repository_root_path, ProcessLaunchInfo(cmd=executable_path, cwd=repository_root_path), "dart", solidlsp_settings
        )

    @classmethod
    def _setup_runtime_dependencies(cls, solidlsp_settings: SolidLSPSettings) -> str:
        deps = RuntimeDependencyCollection(
            [
                RuntimeDependency(
                    id="DartLanguageServer",
                    description="Dart Language Server for Linux (x64)",
                    url="https://storage.googleapis.com/dart-archive/channels/stable/release/3.7.1/sdk/dartsdk-linux-x64-release.zip",
                    platform_id="linux-x64",
                    archive_type="zip",
                    binary_name="dart-sdk/bin/dart",
                ),
                RuntimeDependency(
                    id="DartLanguageServer",
                    description="Dart Language Server for Windows (x64)",
                    url="https://storage.googleapis.com/dart-archive/channels/stable/release/3.7.1/sdk/dartsdk-windows-x64-release.zip",
                    platform_id="win-x64",
                    archive_type="zip",
                    binary_name="dart-sdk/bin/dart.exe",
                ),
                RuntimeDependency(
                    id="DartLanguageServer",
                    description="Dart Language Server for Windows (arm64)",
                    url="https://storage.googleapis.com/dart-archive/channels/stable/release/3.7.1/sdk/dartsdk-windows-arm64-release.zip",
                    platform_id="win-arm64",
                    archive_type="zip",
                    binary_name="dart-sdk/bin/dart.exe",
                ),
                RuntimeDependency(
                    id="DartLanguageServer",
                    description="Dart Language Server for macOS (x64)",
                    url="https://storage.googleapis.com/dart-archive/channels/stable/release/3.7.1/sdk/dartsdk-macos-x64-release.zip",
                    platform_id="osx-x64",
                    archive_type="zip",
                    binary_name="dart-sdk/bin/dart",
                ),
                RuntimeDependency(
                    id="DartLanguageServer",
                    description="Dart Language Server for macOS (arm64)",
                    url="https://storage.googleapis.com/dart-archive/channels/stable/release/3.7.1/sdk/dartsdk-macos-arm64-release.zip",
                    platform_id="osx-arm64",
                    archive_type="zip",
                    binary_name="dart-sdk/bin/dart",
                ),
            ]
        )

        dart_ls_dir = cls.ls_resources_dir(solidlsp_settings)
        dart_executable_path = deps.binary_path(dart_ls_dir)

        if not os.path.exists(dart_executable_path):
            deps.install(dart_ls_dir)

        assert os.path.exists(dart_executable_path)
        os.chmod(dart_executable_path, 0o755)

        return f"{dart_executable_path} language-server --client-id multilspy.dart --client-version 1.2"

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Dart Language Server.
        """
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        initialize_params = {
            "capabilities": {},
            "initializationOptions": {
                "onlyAnalyzeProjectsWithOpenFiles": False,
                "closingLabels": False,
                "outline": False,
                "flutterOutline": False,
                "allowOpenUri": False,
            },
            "trace": "verbose",
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": pathlib.Path(repository_absolute_path).as_uri(),
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }

        return cast(InitializeParams, initialize_params)

    def _start_server(self) -> None:
        """
        Start the language server and yield when the server is ready.
        """

        def execute_client_command_handler(params: dict) -> list:
            return []

        def do_nothing(params: dict) -> None:
            return

        def check_experimental_status(params: dict) -> None:
            pass

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        self.server.on_request("client/registerCapability", do_nothing)
        self.server.on_notification("language/status", do_nothing)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_request("workspace/executeClientCommand", execute_client_command_handler)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)
        self.server.on_notification("language/actionableNotification", do_nothing)
        self.server.on_notification("experimental/serverStatus", check_experimental_status)

        log.info("Starting dart-language-server server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)
        log.debug("Sending initialize request to dart-language-server")
        init_response = self.server.send_request("initialize", initialize_params)  # type: ignore
        log.info(f"Received initialize response from dart-language-server: {init_response}")

        self.server.notify.initialized({})



"""LSP 기본 타입/공용 값 계층을 재노출한다."""

from __future__ import annotations

from .lsp_types_part1 import *  # noqa: F401,F403
from .lsp_types_part2 import *  # noqa: F401,F403
from .lsp_types_part5 import *  # noqa: F401,F403



"""solidlsp에서 사용하는 문자열 유틸을 제공한다."""


class ToStringMixin:
    """객체 문자열 표현을 단순 제공한다."""

    def _tostring_includes(self) -> list[str]:
        """문자열 표현에 포함할 속성 목록을 반환한다."""
        return []

    def __str__(self) -> str:
        """속성 기반 문자열 표현을 생성한다."""
        keys = self._tostring_includes()
        pairs: list[str] = []
        for key in keys:
            pairs.append(f"{key}={getattr(self, key, None)!r}")
        return f"{self.__class__.__name__}({', '.join(pairs)})"



"""L2/L3 파이프라인 워커 전용 컴포넌트."""

from __future__ import annotations

from typing import Callable


class PipelineWorker:
    """보강 워커 책임을 담당하는 전용 서비스."""

    def __init__(
        self,
        *,
        process_enrich_jobs: Callable[[int], int],
        process_enrich_jobs_l2: Callable[[int], int],
        process_enrich_jobs_l3: Callable[[int], int],
    ) -> None:
        """보강 처리 함수를 주입받는다."""
        self._process_enrich_jobs = process_enrich_jobs
        self._process_enrich_jobs_l2 = process_enrich_jobs_l2
        self._process_enrich_jobs_l3 = process_enrich_jobs_l3

    def process_enrich_jobs(self, limit: int) -> int:
        """단일 루프에서 L2/L3 통합 처리를 실행한다."""
        return self._process_enrich_jobs(limit)

    def process_enrich_jobs_l2(self, limit: int) -> int:
        """L2 본문/벡터 보강 처리를 실행한다."""
        return self._process_enrich_jobs_l2(limit)

    def process_enrich_jobs_l3(self, limit: int) -> int:
        """L3 LSP 보강 처리를 실행한다."""
        return self._process_enrich_jobs_l3(limit)



"""MCP stdio 전송 계층(Content-Length/JSONL)을 제공한다."""

from __future__ import annotations

import json
from typing import BinaryIO

MCP_MODE_FRAMED = "content-length"
MCP_MODE_JSONL = "jsonl"
MAX_MESSAGE_SIZE = 10 * 1024 * 1024


class McpTransportParseError(Exception):
    """전송 계층에서 메시지 파싱에 실패했음을 나타낸다."""

    def __init__(self, mode: str, message: str) -> None:
        """오류 모드와 메시지를 저장한다."""
        super().__init__(message)
        self.mode = mode


class McpTransport:
    """MCP wire protocol 읽기/쓰기를 담당한다."""

    def __init__(self, input_stream: BinaryIO, output_stream: BinaryIO, allow_jsonl: bool = True) -> None:
        """입출력 스트림과 모드를 초기화한다."""
        self._input = input_stream
        self._output = output_stream
        self._allow_jsonl = allow_jsonl
        self.default_mode = MCP_MODE_FRAMED

    def read_message(self) -> tuple[dict[str, object], str] | None:
        """다음 MCP 메시지를 읽어 payload와 모드를 반환한다."""
        while True:
            line = self._input.readline()
            if line == b"":
                return None

            line_text = self._decode_utf8_strict(line=line, mode=MCP_MODE_FRAMED).strip()
            if line_text == "":
                continue

            if line_text.startswith("{"):
                if not self._allow_jsonl:
                    continue
                payload = self._parse_json_object(line_text)
                if payload is not None:
                    return payload, MCP_MODE_JSONL
                raise McpTransportParseError(mode=MCP_MODE_JSONL, message="parse error")

            if line_text.lower().startswith("content-length:"):
                content_length = self._parse_headers(first_line=line_text)
                if content_length is None:
                    continue
                if content_length <= 0 or content_length > MAX_MESSAGE_SIZE:
                    continue

                body = self._input.read(content_length)
                if len(body) < content_length:
                    return None
                body_text = self._decode_utf8_strict(line=body, mode=MCP_MODE_FRAMED)
                payload = self._parse_json_object(body_text)
                if payload is not None:
                    return payload, MCP_MODE_FRAMED
                raise McpTransportParseError(mode=MCP_MODE_FRAMED, message="parse error")

    def write_message(self, message: dict[str, object], mode: str | None = None) -> None:
        """모드에 맞춰 MCP 응답 메시지를 출력한다."""
        encoded = json.dumps(message, ensure_ascii=False).encode("utf-8")
        selected_mode = mode if mode is not None else self.default_mode
        if selected_mode == MCP_MODE_JSONL:
            self._output.write(encoded + b"\n")
            self._output.flush()
            return

        header = f"Content-Length: {len(encoded)}\r\n\r\n".encode("ascii")
        self._output.write(header + encoded)
        self._output.flush()

    def _parse_headers(self, first_line: str) -> int | None:
        """헤더 블록에서 content-length를 파싱한다."""
        headers: dict[str, str] = {}
        first_key, first_value = self._split_header(first_line)
        if first_key is None or first_value is None:
            return None
        headers[first_key] = first_value

        while True:
            line = self._input.readline()
            if line == b"":
                break
            header_line = self._decode_utf8_strict(line=line, mode=MCP_MODE_FRAMED).strip()
            if header_line == "":
                break
            key, value = self._split_header(header_line)
            if key is None or value is None:
                continue
            headers[key] = value

        raw_length = headers.get("content-length")
        if raw_length is None:
            return None
        try:
            return int(raw_length)
        except ValueError:
            return None

    @staticmethod
    def _split_header(line: str) -> tuple[str | None, str | None]:
        """단일 헤더 라인을 key/value로 분리한다."""
        if ":" not in line:
            return None, None
        key, value = line.split(":", 1)
        return key.strip().lower(), value.strip()

    @staticmethod
    def _parse_json_object(text: str) -> dict[str, object] | None:
        """JSON 문자열을 dict payload로 파싱한다."""
        try:
            parsed = json.loads(text)
        except json.JSONDecodeError:
            return None
        if not isinstance(parsed, dict):
            return None
        normalized: dict[str, object] = {}
        for key, value in parsed.items():
            if isinstance(key, str):
                normalized[key] = value
        return normalized

    @staticmethod
    def _decode_utf8_strict(line: bytes, mode: str) -> str:
        """UTF-8 디코드 실패를 침묵하지 않고 전송 계층 오류로 변환한다."""
        try:
            return line.decode("utf-8")
        except UnicodeDecodeError as exc:
            raise McpTransportParseError(mode=mode, message=f"invalid utf-8 payload: {exc}") from exc
